mooniswap_abi = [{'inputs': [{'internalType': 'contract IERC20',
                              'name': '_token0',
                              'type': 'address'},
                             {'internalType': 'contract IERC20', 'name': '_token1', 'type': 'address'},
                             {'internalType': 'string', 'name': 'name', 'type': 'string'},
                             {'internalType': 'string', 'name': 'symbol', 'type': 'string'},
                             {'internalType': 'contract IMooniswapFactoryGovernance',
                              'name': '_mooniswapFactoryGovernance',
                              'type': 'address'}],
                  'stateMutability': 'nonpayable',
                  'type': 'constructor'},
                 {'anonymous': False,
                  'inputs': [{'indexed': True,
                              'internalType': 'address',
                              'name': 'owner',
                              'type': 'address'},
                             {'indexed': True,
                              'internalType': 'address',
                              'name': 'spender',
                              'type': 'address'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'value',
                              'type': 'uint256'}],
                  'name': 'Approval',
                  'type': 'event'},
                 {'anonymous': False,
                  'inputs': [{'indexed': True,
                              'internalType': 'address',
                              'name': 'user',
                              'type': 'address'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'decayPeriod',
                              'type': 'uint256'},
                             {'indexed': False,
                              'internalType': 'bool',
                              'name': 'isDefault',
                              'type': 'bool'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'amount',
                              'type': 'uint256'}],
                  'name': 'DecayPeriodVoteUpdate',
                  'type': 'event'},
                 {'anonymous': False,
                  'inputs': [{'indexed': True,
                              'internalType': 'address',
                              'name': 'sender',
                              'type': 'address'},
                             {'indexed': True,
                              'internalType': 'address',
                              'name': 'receiver',
                              'type': 'address'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'share',
                              'type': 'uint256'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'token0Amount',
                              'type': 'uint256'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'token1Amount',
                              'type': 'uint256'}],
                  'name': 'Deposited',
                  'type': 'event'},
                 {'anonymous': False,
                  'inputs': [{'indexed': False,
                              'internalType': 'string',
                              'name': 'reason',
                              'type': 'string'}],
                  'name': 'Error',
                  'type': 'event'},
                 {'anonymous': False,
                  'inputs': [{'indexed': True,
                              'internalType': 'address',
                              'name': 'user',
                              'type': 'address'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'fee',
                              'type': 'uint256'},
                             {'indexed': False,
                              'internalType': 'bool',
                              'name': 'isDefault',
                              'type': 'bool'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'amount',
                              'type': 'uint256'}],
                  'name': 'FeeVoteUpdate',
                  'type': 'event'},
                 {'anonymous': False,
                  'inputs': [{'indexed': True,
                              'internalType': 'address',
                              'name': 'previousOwner',
                              'type': 'address'},
                             {'indexed': True,
                              'internalType': 'address',
                              'name': 'newOwner',
                              'type': 'address'}],
                  'name': 'OwnershipTransferred',
                  'type': 'event'},
                 {'anonymous': False,
                  'inputs': [{'indexed': True,
                              'internalType': 'address',
                              'name': 'user',
                              'type': 'address'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'slippageFee',
                              'type': 'uint256'},
                             {'indexed': False,
                              'internalType': 'bool',
                              'name': 'isDefault',
                              'type': 'bool'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'amount',
                              'type': 'uint256'}],
                  'name': 'SlippageFeeVoteUpdate',
                  'type': 'event'},
                 {'anonymous': False,
                  'inputs': [{'indexed': True,
                              'internalType': 'address',
                              'name': 'sender',
                              'type': 'address'},
                             {'indexed': True,
                              'internalType': 'address',
                              'name': 'receiver',
                              'type': 'address'},
                             {'indexed': True,
                              'internalType': 'address',
                              'name': 'srcToken',
                              'type': 'address'},
                             {'indexed': False,
                              'internalType': 'address',
                              'name': 'dstToken',
                              'type': 'address'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'amount',
                              'type': 'uint256'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'result',
                              'type': 'uint256'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'srcAdditionBalance',
                              'type': 'uint256'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'dstRemovalBalance',
                              'type': 'uint256'},
                             {'indexed': False,
                              'internalType': 'address',
                              'name': 'referral',
                              'type': 'address'}],
                  'name': 'Swapped',
                  'type': 'event'},
                 {'anonymous': False,
                  'inputs': [{'indexed': False,
                              'internalType': 'uint256',
                              'name': 'srcBalance',
                              'type': 'uint256'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'dstBalance',
                              'type': 'uint256'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'fee',
                              'type': 'uint256'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'slippageFee',
                              'type': 'uint256'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'referralShare',
                              'type': 'uint256'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'governanceShare',
                              'type': 'uint256'}],
                  'name': 'Sync',
                  'type': 'event'},
                 {'anonymous': False,
                  'inputs': [{'indexed': True,
                              'internalType': 'address',
                              'name': 'from',
                              'type': 'address'},
                             {'indexed': True,
                              'internalType': 'address',
                              'name': 'to',
                              'type': 'address'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'value',
                              'type': 'uint256'}],
                  'name': 'Transfer',
                  'type': 'event'},
                 {'anonymous': False,
                  'inputs': [{'indexed': True,
                              'internalType': 'address',
                              'name': 'sender',
                              'type': 'address'},
                             {'indexed': True,
                              'internalType': 'address',
                              'name': 'receiver',
                              'type': 'address'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'share',
                              'type': 'uint256'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'token0Amount',
                              'type': 'uint256'},
                             {'indexed': False,
                              'internalType': 'uint256',
                              'name': 'token1Amount',
                              'type': 'uint256'}],
                  'name': 'Withdrawn',
                  'type': 'event'},
                 {'inputs': [{'internalType': 'address', 'name': 'owner', 'type': 'address'},
                             {'internalType': 'address', 'name': 'spender', 'type': 'address'}],
                  'name': 'allowance',
                  'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'address', 'name': 'spender', 'type': 'address'},
                             {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'}],
                  'name': 'approve',
                  'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
                  'stateMutability': 'nonpayable',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'address',
                              'name': 'account',
                              'type': 'address'}],
                  'name': 'balanceOf',
                  'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'decayPeriod',
                  'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'uint256', 'name': 'vote', 'type': 'uint256'}],
                  'name': 'decayPeriodVote',
                  'outputs': [],
                  'stateMutability': 'nonpayable',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'address', 'name': 'user', 'type': 'address'}],
                  'name': 'decayPeriodVotes',
                  'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'decimals',
                  'outputs': [{'internalType': 'uint8', 'name': '', 'type': 'uint8'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'address', 'name': 'spender', 'type': 'address'},
                             {'internalType': 'uint256', 'name': 'subtractedValue', 'type': 'uint256'}],
                  'name': 'decreaseAllowance',
                  'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
                  'stateMutability': 'nonpayable',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'uint256[2]',
                              'name': 'maxAmounts',
                              'type': 'uint256[2]'},
                             {'internalType': 'uint256[2]', 'name': 'minAmounts', 'type': 'uint256[2]'}],
                  'name': 'deposit',
                  'outputs': [{'internalType': 'uint256',
                               'name': 'fairSupply',
                               'type': 'uint256'},
                              {'internalType': 'uint256[2]',
                               'name': 'receivedAmounts',
                               'type': 'uint256[2]'}],
                  'stateMutability': 'payable',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'uint256[2]',
                              'name': 'maxAmounts',
                              'type': 'uint256[2]'},
                             {'internalType': 'uint256[2]', 'name': 'minAmounts', 'type': 'uint256[2]'},
                             {'internalType': 'address', 'name': 'target', 'type': 'address'}],
                  'name': 'depositFor',
                  'outputs': [{'internalType': 'uint256',
                               'name': 'fairSupply',
                               'type': 'uint256'},
                              {'internalType': 'uint256[2]',
                               'name': 'receivedAmounts',
                               'type': 'uint256[2]'}],
                  'stateMutability': 'payable',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'discardDecayPeriodVote',
                  'outputs': [],
                  'stateMutability': 'nonpayable',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'discardFeeVote',
                  'outputs': [],
                  'stateMutability': 'nonpayable',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'discardSlippageFeeVote',
                  'outputs': [],
                  'stateMutability': 'nonpayable',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'fee',
                  'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'uint256', 'name': 'vote', 'type': 'uint256'}],
                  'name': 'feeVote',
                  'outputs': [],
                  'stateMutability': 'nonpayable',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'address', 'name': 'user', 'type': 'address'}],
                  'name': 'feeVotes',
                  'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'contract IERC20',
                              'name': 'token',
                              'type': 'address'}],
                  'name': 'getBalanceForAddition',
                  'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'contract IERC20',
                              'name': 'token',
                              'type': 'address'}],
                  'name': 'getBalanceForRemoval',
                  'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'contract IERC20',
                              'name': 'src',
                              'type': 'address'},
                             {'internalType': 'contract IERC20', 'name': 'dst', 'type': 'address'},
                             {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'}],
                  'name': 'getReturn',
                  'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'getTokens',
                  'outputs': [{'internalType': 'contract IERC20[]',
                               'name': 'tokens',
                               'type': 'address[]'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'address', 'name': 'spender', 'type': 'address'},
                             {'internalType': 'uint256', 'name': 'addedValue', 'type': 'uint256'}],
                  'name': 'increaseAllowance',
                  'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
                  'stateMutability': 'nonpayable',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'mooniswapFactoryGovernance',
                  'outputs': [{'internalType': 'contract IMooniswapFactoryGovernance',
                               'name': '',
                               'type': 'address'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'name',
                  'outputs': [{'internalType': 'string', 'name': '', 'type': 'string'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'owner',
                  'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'renounceOwnership',
                  'outputs': [],
                  'stateMutability': 'nonpayable',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'contract IERC20',
                              'name': 'token',
                              'type': 'address'},
                             {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'}],
                  'name': 'rescueFunds',
                  'outputs': [],
                  'stateMutability': 'nonpayable',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'contract IMooniswapFactoryGovernance',
                              'name': 'newMooniswapFactoryGovernance',
                              'type': 'address'}],
                  'name': 'setMooniswapFactoryGovernance',
                  'outputs': [],
                  'stateMutability': 'nonpayable',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'slippageFee',
                  'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'uint256', 'name': 'vote', 'type': 'uint256'}],
                  'name': 'slippageFeeVote',
                  'outputs': [],
                  'stateMutability': 'nonpayable',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'address', 'name': 'user', 'type': 'address'}],
                  'name': 'slippageFeeVotes',
                  'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'contract IERC20',
                              'name': 'src',
                              'type': 'address'},
                             {'internalType': 'contract IERC20', 'name': 'dst', 'type': 'address'},
                             {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'},
                             {'internalType': 'uint256', 'name': 'minReturn', 'type': 'uint256'},
                             {'internalType': 'address', 'name': 'referral', 'type': 'address'}],
                  'name': 'swap',
                  'outputs': [{'internalType': 'uint256',
                               'name': 'result',
                               'type': 'uint256'}],
                  'stateMutability': 'payable',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'contract IERC20',
                              'name': 'src',
                              'type': 'address'},
                             {'internalType': 'contract IERC20', 'name': 'dst', 'type': 'address'},
                             {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'},
                             {'internalType': 'uint256', 'name': 'minReturn', 'type': 'uint256'},
                             {'internalType': 'address', 'name': 'referral', 'type': 'address'},
                             {'internalType': 'address payable', 'name': 'receiver', 'type': 'address'}],
                  'name': 'swapFor',
                  'outputs': [{'internalType': 'uint256',
                               'name': 'result',
                               'type': 'uint256'}],
                  'stateMutability': 'payable',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'symbol',
                  'outputs': [{'internalType': 'string', 'name': '', 'type': 'string'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'token0',
                  'outputs': [{'internalType': 'contract IERC20',
                               'name': '',
                               'type': 'address'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'token1',
                  'outputs': [{'internalType': 'contract IERC20',
                               'name': '',
                               'type': 'address'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'uint256', 'name': 'i', 'type': 'uint256'}],
                  'name': 'tokens',
                  'outputs': [{'internalType': 'contract IERC20',
                               'name': '',
                               'type': 'address'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'totalSupply',
                  'outputs': [{'internalType': 'uint256', 'name': '', 'type': 'uint256'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'address',
                              'name': 'recipient',
                              'type': 'address'},
                             {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'}],
                  'name': 'transfer',
                  'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
                  'stateMutability': 'nonpayable',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'address', 'name': 'sender', 'type': 'address'},
                             {'internalType': 'address', 'name': 'recipient', 'type': 'address'},
                             {'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'}],
                  'name': 'transferFrom',
                  'outputs': [{'internalType': 'bool', 'name': '', 'type': 'bool'}],
                  'stateMutability': 'nonpayable',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'address',
                              'name': 'newOwner',
                              'type': 'address'}],
                  'name': 'transferOwnership',
                  'outputs': [],
                  'stateMutability': 'nonpayable',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'contract IERC20',
                              'name': '',
                              'type': 'address'}],
                  'name': 'virtualBalancesForAddition',
                  'outputs': [{'internalType': 'uint216',
                               'name': 'balance',
                               'type': 'uint216'},
                              {'internalType': 'uint40', 'name': 'time', 'type': 'uint40'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'contract IERC20',
                              'name': '',
                              'type': 'address'}],
                  'name': 'virtualBalancesForRemoval',
                  'outputs': [{'internalType': 'uint216',
                               'name': 'balance',
                               'type': 'uint216'},
                              {'internalType': 'uint40', 'name': 'time', 'type': 'uint40'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'virtualDecayPeriod',
                  'outputs': [{'internalType': 'uint104', 'name': '', 'type': 'uint104'},
                              {'internalType': 'uint104', 'name': '', 'type': 'uint104'},
                              {'internalType': 'uint48', 'name': '', 'type': 'uint48'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'virtualFee',
                  'outputs': [{'internalType': 'uint104', 'name': '', 'type': 'uint104'},
                              {'internalType': 'uint104', 'name': '', 'type': 'uint104'},
                              {'internalType': 'uint48', 'name': '', 'type': 'uint48'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [],
                  'name': 'virtualSlippageFee',
                  'outputs': [{'internalType': 'uint104', 'name': '', 'type': 'uint104'},
                              {'internalType': 'uint104', 'name': '', 'type': 'uint104'},
                              {'internalType': 'uint48', 'name': '', 'type': 'uint48'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'contract IERC20',
                              'name': '',
                              'type': 'address'}],
                  'name': 'volumes',
                  'outputs': [{'internalType': 'uint128',
                               'name': 'confirmed',
                               'type': 'uint128'},
                              {'internalType': 'uint128', 'name': 'result', 'type': 'uint128'}],
                  'stateMutability': 'view',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'},
                             {'internalType': 'uint256[]', 'name': 'minReturns', 'type': 'uint256[]'}],
                  'name': 'withdraw',
                  'outputs': [{'internalType': 'uint256[2]',
                               'name': 'withdrawnAmounts',
                               'type': 'uint256[2]'}],
                  'stateMutability': 'nonpayable',
                  'type': 'function'},
                 {'inputs': [{'internalType': 'uint256', 'name': 'amount', 'type': 'uint256'},
                             {'internalType': 'uint256[]', 'name': 'minReturns', 'type': 'uint256[]'},
                             {'internalType': 'address payable', 'name': 'target', 'type': 'address'}],
                  'name': 'withdrawFor',
                  'outputs': [{'internalType': 'uint256[2]',
                               'name': 'withdrawnAmounts',
                               'type': 'uint256[2]'}],
                  'stateMutability': 'nonpayable',
                  'type': 'function'}]



import logging

from django.test import TestCase

from eth_account import Account

from gnosis.safe import Safe

from ..proxy_factory import ProxyFactory
from .safe_test_case import SafeTestCaseMixin
from .utils import generate_salt_nonce

logger = logging.getLogger(__name__)


class TestProxyFactory(SafeTestCaseMixin, TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.proxy_factory = ProxyFactory(cls.proxy_factory_contract_address, cls.ethereum_client)

    def test_check_proxy_code(self):
        proxy_contract_address = self.deploy_test_safe().safe_address
        self.assertTrue(self.proxy_factory.check_proxy_code(proxy_contract_address))

        ethereum_tx_sent = Safe.deploy_master_contract(self.ethereum_client, self.ethereum_test_account)
        self.assertFalse(self.proxy_factory.check_proxy_code(ethereum_tx_sent.contract_address))

        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract(self.ethereum_test_account,
                                                                    Account.create().address)
        self.assertTrue(self.proxy_factory.check_proxy_code(ethereum_tx_sent.contract_address))

    def test_deploy_proxy_contract(self):
        s = 15
        owners = [Account.create().address for _ in range(2)]
        threshold = 2
        payment_token = None
        safe_creation_tx = Safe.build_safe_creation_tx(self.ethereum_client, self.safe_contract_V0_0_1_address,
                                                       s, owners, threshold, self.gas_price, payment_token,
                                                       payment_receiver=self.ethereum_test_account.address)
        # Send ether for safe deploying costs
        self.send_tx({
            'to': safe_creation_tx.safe_address,
            'value': safe_creation_tx.payment
        }, self.ethereum_test_account)

        proxy_factory = ProxyFactory(self.proxy_factory_contract_address, self.ethereum_client)
        ethereum_tx_sent = proxy_factory.deploy_proxy_contract(self.ethereum_test_account,
                                                               safe_creation_tx.master_copy,
                                                               safe_creation_tx.safe_setup_data,
                                                               safe_creation_tx.gas,
                                                               gas_price=self.gas_price)
        receipt = self.ethereum_client.get_transaction_receipt(ethereum_tx_sent.tx_hash, timeout=20)
        self.assertEqual(receipt.status, 1)
        safe = Safe(ethereum_tx_sent.contract_address, self.ethereum_client)
        self.assertEqual(safe.retrieve_master_copy_address(), safe_creation_tx.master_copy)
        self.assertEqual(set(safe.retrieve_owners()), set(owners))

    def test_deploy_proxy_contract_with_nonce(self):
        salt_nonce = generate_salt_nonce()
        owners = [Account.create().address for _ in range(2)]
        threshold = 2
        payment_token = None
        safe_create2_tx = Safe.build_safe_create2_tx(self.ethereum_client, self.safe_contract_address,
                                                     self.proxy_factory_contract_address, salt_nonce,
                                                     owners, threshold, self.gas_price, payment_token)
        # Send ether for safe deploying costs
        self.send_tx({
            'to': safe_create2_tx.safe_address,
            'value': safe_create2_tx.payment
        }, self.ethereum_test_account)

        proxy_factory = ProxyFactory(self.proxy_factory_contract_address, self.ethereum_client)
        ethereum_tx_sent = proxy_factory.deploy_proxy_contract_with_nonce(self.ethereum_test_account,
                                                                          safe_create2_tx.master_copy_address,
                                                                          safe_create2_tx.safe_setup_data,
                                                                          salt_nonce,
                                                                          safe_create2_tx.gas,
                                                                          gas_price=self.gas_price)
        receipt = self.ethereum_client.get_transaction_receipt(ethereum_tx_sent.tx_hash, timeout=20)
        self.assertEqual(receipt.status, 1)
        safe = Safe(ethereum_tx_sent.contract_address, self.ethereum_client)
        self.assertEqual(ethereum_tx_sent.contract_address, safe_create2_tx.safe_address)
        self.assertEqual(set(safe.retrieve_owners()), set(owners))
        self.assertEqual(safe.retrieve_master_copy_address(), safe_create2_tx.master_copy_address)

    def test_get_proxy_runtime_code(self):
        self.assertGreater(len(self.proxy_factory.get_proxy_runtime_code()), 4)



import functools
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

import requests
from eth_abi.exceptions import DecodingError
from eth_abi.packed import encode_abi_packed
from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import Contract
from web3.exceptions import BadFunctionCallOutput

from .. import EthereumClient
from ..constants import NULL_ADDRESS
from ..contracts import (get_erc20_contract, get_kyber_network_proxy_contract,
                         get_uniswap_factory_contract,
                         get_uniswap_v2_factory_contract,
                         get_uniswap_v2_pair_contract,
                         get_uniswap_v2_router_contract)
from ..ethereum_client import EthereumNetwork
from .abis.aave_abis import AAVE_ATOKEN_ABI
from .abis.balancer_abis import balancer_pool_abi
from .abis.mooniswap_abis import mooniswap_abi
from .abis.zerion_abis import ZERION_TOKEN_ADAPTER_ABI

try:
    from functools import cached_property
except ImportError:
    from cached_property import cached_property


logger = logging.getLogger(__name__)


@dataclass
class UnderlyingToken:
    address: ChecksumAddress
    quantity: int


class OracleException(Exception):
    pass


class InvalidPriceFromOracle(OracleException):
    pass


class CannotGetPriceFromOracle(OracleException):
    pass


class PriceOracle(ABC):
    @abstractmethod
    def get_price(self, *args) -> float:
        pass


class PricePoolOracle(ABC):
    @abstractmethod
    def get_pool_token_price(self, pool_token_address: ChecksumAddress) -> float:
        pass


class UsdPricePoolOracle(ABC):
    @abstractmethod
    def get_pool_usd_token_price(self, pool_token_address: ChecksumAddress) -> float:
        pass


class ComposedPriceOracle(ABC):
    @abstractmethod
    def get_underlying_tokens(self, *args) -> List[Tuple[UnderlyingToken]]:
        pass


class KyberOracle(PriceOracle):
    # This is the `tokenAddress` they use for ETH ¯\_(ツ)_/¯
    ETH_TOKEN_ADDRESS = '0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE'
    ADDRESSES = {
        EthereumNetwork.MAINNET: '0x9AAb3f75489902f3a48495025729a0AF77d4b11e',
        EthereumNetwork.RINKEBY: '0x0d5371e5EE23dec7DF251A8957279629aa79E9C5',
        EthereumNetwork.ROPSTEN: '0xd719c34261e099Fdb33030ac8909d5788D3039C4',
        EthereumNetwork.KOVAN: '0xc153eeAD19e0DBbDb3462Dcc2B703cC6D738A37c',
    }

    def __init__(self, ethereum_client: EthereumClient,
                 kyber_network_proxy_address: Optional[str] = None):
        """
        :param ethereum_client:
        :param kyber_network_proxy_address: https://developer.kyber.network/docs/MainnetEnvGuide/#contract-addresses
        """
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self._kyber_network_proxy_address = kyber_network_proxy_address

    @cached_property
    def kyber_network_proxy_address(self):
        if self._kyber_network_proxy_address:
            return self._kyber_network_proxy_address
        return self.ADDRESSES.get(self.ethereum_client.get_network(),
                                  self.ADDRESSES.get(EthereumNetwork.MAINNET))  # By default return Mainnet address

    @cached_property
    def kyber_network_proxy_contract(self):
        return get_kyber_network_proxy_contract(self.w3, self.kyber_network_proxy_address)

    def get_price(self, token_address_1: str, token_address_2: str = ETH_TOKEN_ADDRESS) -> float:
        if token_address_1 == token_address_2:
            return 1.
        try:
            # Get decimals for token, estimation will be more accurate
            decimals = self.ethereum_client.erc20.get_decimals(token_address_1)
            token_unit = int(10 ** decimals)
            expected_rate, _ = self.kyber_network_proxy_contract.functions.getExpectedRate(token_address_1,
                                                                                           token_address_2,
                                                                                           int(token_unit)).call()

            price = expected_rate / 1e18

            if price <= 0.:
                # Try again the opposite
                expected_rate, _ = self.kyber_network_proxy_contract.functions.getExpectedRate(token_address_2,
                                                                                               token_address_1,
                                                                                               int(token_unit)).call()
                price = (token_unit / expected_rate) if expected_rate else 0

            if price <= 0.:
                error_message = f'price={price} <= 0 from kyber-network-proxy={self.kyber_network_proxy_address} ' \
                                f'for token-1={token_address_1} to token-2={token_address_2}'
                logger.warning(error_message)
                raise InvalidPriceFromOracle(error_message)
            return price
        except (ValueError, BadFunctionCallOutput, DecodingError) as e:
            error_message = f'Cannot get price from kyber-network-proxy={self.kyber_network_proxy_address} ' \
                            f'for token-1={token_address_1} to token-2={token_address_2}'
            logger.warning(error_message)
            raise CannotGetPriceFromOracle(error_message) from e


class UniswapOracle(PriceOracle):
    ADDRESSES = {
        EthereumNetwork.MAINNET: '0xc0a47dFe034B400B47bDaD5FecDa2621de6c4d95',
        EthereumNetwork.RINKEBY: '0xf5D915570BC477f9B8D6C0E980aA81757A3AaC36',
        EthereumNetwork.ROPSTEN: '0x9c83dCE8CA20E9aAF9D3efc003b2ea62aBC08351',
        EthereumNetwork.KOVAN: '0xD3E51Ef092B2845f10401a0159B2B96e8B6c3D30',
        EthereumNetwork.GOERLI: '0x6Ce570d02D73d4c384b46135E87f8C592A8c86dA',
    }

    def __init__(self, ethereum_client: EthereumClient, uniswap_factory_address: Optional[str] = None):
        """
        :param ethereum_client:
        :param uniswap_factory_address: https://docs.uniswap.io/frontend-integration/connect-to-uniswap#factory-contract
        """
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self._uniswap_factory_address = uniswap_factory_address

    @cached_property
    def uniswap_factory_address(self):
        if self._uniswap_factory_address:
            return self._uniswap_factory_address
        return self.ADDRESSES.get(self.ethereum_client.get_network(),
                                  self.ADDRESSES.get(EthereumNetwork.MAINNET))

    @cached_property
    def uniswap_factory(self):
        return get_uniswap_factory_contract(self.w3, self.uniswap_factory_address)

    @functools.lru_cache(maxsize=None)
    def get_uniswap_exchange(self, token_address: str) -> str:
        return self.uniswap_factory.functions.getExchange(token_address).call()

    def _get_balances_without_batching(self, uniswap_exchange_address: str, token_address: str):
        balance = self.ethereum_client.get_balance(uniswap_exchange_address)
        token_decimals = self.ethereum_client.erc20.get_decimals(token_address)
        token_balance = self.ethereum_client.erc20.get_balance(uniswap_exchange_address, token_address)
        return balance, token_decimals, token_balance

    def _get_balances_using_batching(self, uniswap_exchange_address: str, token_address: str):
        # Use batching instead
        payload_balance = {'jsonrpc': '2.0',
                           'method': 'eth_getBalance',
                           'params': [uniswap_exchange_address, 'latest'],
                           'id': 0}
        erc20 = get_erc20_contract(self.w3, token_address)
        params = {'gas': 0, 'gasPrice': 0}
        decimals_data = erc20.functions.decimals().buildTransaction(params)['data']
        token_balance_data = erc20.functions.balanceOf(uniswap_exchange_address).buildTransaction(params)['data']
        datas = [decimals_data, token_balance_data]
        payload_calls = [{'id': i + 1, 'jsonrpc': '2.0', 'method': 'eth_call',
                          'params': [{'to': token_address, 'data': data}, 'latest']}
                         for i, data in enumerate(datas)]
        payloads = [payload_balance] + payload_calls
        r = requests.post(self.ethereum_client.ethereum_node_url, json=payloads)
        if not r.ok:
            raise CannotGetPriceFromOracle(f'Error from node with url={self.ethereum_client.ethereum_node_url}')

        results = []
        for result in r.json():
            if 'result' not in result:
                raise CannotGetPriceFromOracle(result['error'])
            else:
                results.append(HexBytes(result['result']))

        balance = int(results[0].hex(), 16)
        token_decimals = self.ethereum_client.w3.codec.decode_single('uint8', results[1])
        token_balance = self.ethereum_client.w3.codec.decode_single('uint256', results[2])
        return balance, token_decimals, token_balance

    def get_price(self, token_address: str) -> float:
        try:
            uniswap_exchange_address = self.get_uniswap_exchange(token_address)
            if uniswap_exchange_address == NULL_ADDRESS:
                raise ValueError
        except (ValueError, BadFunctionCallOutput, DecodingError) as e:
            error_message = f'Non existing uniswap exchange for token={token_address}'
            logger.warning(error_message)
            raise CannotGetPriceFromOracle(error_message) from e

        try:
            balance, token_decimals, token_balance = self._get_balances_using_batching(uniswap_exchange_address,
                                                                                       token_address)
            # Check liquidity. Require at least 2 ether to be on the pool
            if balance / 1e18 < 2:
                raise CannotGetPriceFromOracle(f'Not enough liquidity for token={token_address}')

            price = balance / token_balance / 10**(18 - token_decimals)
            if price <= 0.:
                error_message = f'price={price} <= 0 from uniswap-factory={uniswap_exchange_address} ' \
                                f'for token={token_address}'
                logger.warning(error_message)
                raise InvalidPriceFromOracle(error_message)
            return price
        except (ValueError, ZeroDivisionError, BadFunctionCallOutput, DecodingError) as e:
            error_message = f'Cannot get token balance for token={token_address}'
            logger.warning(error_message)
            raise CannotGetPriceFromOracle(error_message) from e


class UniswapV2Oracle(PricePoolOracle, PriceOracle):
    # Pair init code is keccak(getCode(UniswapV2Pair))
    pair_init_code = HexBytes('0x96e8ac4277198ff8b6f785478aa9a39f403cb768dd02cbee326c3e7da348845f')
    router_address = '0x7a250d5630B4cF539739dF2C5dAcb4c659F2488D'  # Same address on every network

    def __init__(self, ethereum_client: EthereumClient,
                 router_address: Optional[str] = None):
        """
        :param ethereum_client:
        :param router_address: https://uniswap.org/docs/v2/smart-contracts/router02/
        """
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self.router_address: str = router_address or self.router_address
        self.router = get_uniswap_v2_router_contract(ethereum_client.w3, self.router_address)
        self._decimals_cache: Dict[str, int] = {}

    @cached_property
    def factory(self):
        return get_uniswap_v2_factory_contract(self.ethereum_client.w3, self.factory_address)

    @cached_property
    def factory_address(self) -> str:
        """
        :return: Uniswap factory checksummed address
        :raises: BadFunctionCallOutput: If router contract is not deployed
        """
        return self.router.functions.factory().call()

    @cached_property
    def weth_address(self) -> str:
        """
        :return: Wrapped ether checksummed address
        :raises: BadFunctionCallOutput: If router contract is not deployed
        """
        return self.router.functions.WETH().call()

    @functools.lru_cache(maxsize=None)
    def get_pair_address(self, token_address: str, token_address_2: str) -> Optional[str]:
        """
        Get uniswap pair address. `token_address` and `token_address_2` are interchangeable.
        https://uniswap.org/docs/v2/smart-contracts/factory/

        :param token_address:
        :param token_address_2:
        :return: Address of the pair for `token_address` and `token_address_2`, if it has been created, else `None`.
        """
        # Token order does not matter for getting pair, just for creating or querying PairCreated event
        pair_address = self.factory.functions.getPair(token_address, token_address_2).call()
        if pair_address == NULL_ADDRESS:
            return None
        return pair_address

    def calculate_pair_address(self, token_address: str, token_address_2: str):
        """
        Calculate pair address without querying blockchain.
        https://uniswap.org/docs/v2/smart-contract-integration/getting-pair-addresses/#docs-header

        :param token_address:
        :param token_address_2:
        :return: Checksummed address for token pair. It could be not created yet
        """
        if token_address.lower() > token_address_2.lower():
            token_address, token_address_2 = token_address_2, token_address
        salt = Web3.keccak(encode_abi_packed(['address', 'address'],
                                             [token_address, token_address_2]))
        address = Web3.keccak(
            encode_abi_packed(['bytes', 'address', 'bytes', 'bytes'],
                              [HexBytes('ff'), self.factory_address, salt, self.pair_init_code]
                              ))[-20:]
        return Web3.toChecksumAddress(address)

    def get_decimals(self, token_address: str, token_address_2: str) -> Tuple[int, int]:
        if not (token_address in self._decimals_cache and token_address_2 in self._decimals_cache):
            decimals_1, decimals_2 = self.ethereum_client.batch_call(
                [
                    get_erc20_contract(self.w3, token_address).functions.decimals(),
                    get_erc20_contract(self.w3, token_address_2).functions.decimals(),
                ])
            self._decimals_cache[token_address] = decimals_1
            self._decimals_cache[token_address_2] = decimals_2
        return self._decimals_cache[token_address], self._decimals_cache[token_address_2]

    def get_reserves(self, pair_address: str) -> Tuple[int, int]:
        """
        Returns the
        Also returns the block.timestamp (mod 2**32) of the last block during which an interaction occured for the pair.
        https://uniswap.org/docs/v2/smart-contracts/pair/
        :return: Reserves of `token_address` and `token_address_2` used to price trades and distribute liquidity.
        """
        pair_contract = get_uniswap_v2_pair_contract(self.ethereum_client.w3, pair_address)
        # Reserves return token_1 reserves, token_2 reserves and block timestamp (mod 2**32) of last interaction
        reserves_1, reserves_2, _ = pair_contract.functions.getReserves().call()
        return reserves_1, reserves_2

    def get_price(self, token_address: str, token_address_2: Optional[str] = None) -> float:
        # These lines only make sense when `get_pair_address` is used. `calculate_pair_address` will always return
        # an address, even it that exchange is not deployed

        # pair_address = self.get_pair_address(token_address, token_address_2)
        # if not pair_address:
        #    error_message = f'Non existing uniswap V2 exchange for token={token_address}'
        #    logger.warning(error_message)
        #    raise CannotGetPriceFromOracle(error_message)
        try:
            token_address_2 = token_address_2 if token_address_2 else self.weth_address
            if token_address == token_address_2:
                return 1.
            pair_address = self.calculate_pair_address(token_address, token_address_2)
            # Tokens are sorted, so token_1 < token_2
            reserves_1, reserves_2 = self.get_reserves(pair_address)
            decimals_1, decimals_2 = self.get_decimals(token_address, token_address_2)
            if token_address.lower() > token_address_2.lower():
                reserves_2, reserves_1 = reserves_1, reserves_2

            # Check liquidity. Require at least 2 units of every token to be on the pool
            if reserves_1 / 10 ** decimals_1 < 2 or reserves_2 / 10 ** decimals_2 < 2:
                raise CannotGetPriceFromOracle(f'Not enough liquidity for pair token_1={token_address} '
                                               f'token_2={token_address_2}')
            decimals_normalized_reserves_1 = reserves_1 * 10 ** decimals_2
            decimals_normalized_reserves_2 = reserves_2 * 10 ** decimals_1

            return decimals_normalized_reserves_2 / decimals_normalized_reserves_1
        except (ValueError, ZeroDivisionError, BadFunctionCallOutput, DecodingError) as e:
            error_message = f'Cannot get uniswap v2 price for pair token_1={token_address} ' \
                            f'token_2={token_address_2}'
            logger.warning(error_message)
            raise CannotGetPriceFromOracle(error_message) from e

    def get_price_without_exception(self, token_address: str, token_address_2: Optional[str] = None) -> float:
        """
        :param token_address:
        :param token_address_2:
        :return: Call `get_price`, return 0. instead on an exception if there's any issue
        """
        try:
            return self.get_price(token_address, token_address_2=token_address_2)
        except CannotGetPriceFromOracle:
            return 0.

    def get_pool_token_price(self, pool_token_address: ChecksumAddress) -> float:
        """
        Estimate pool token price based on its components

        :param pool_token_address:
        :return: Pool token eth price per unit (total pool token supply / 1e18)
        :raises: CannotGetPriceFromOracle
        """
        try:
            pair_contract = get_uniswap_v2_pair_contract(self.ethereum_client.w3, pool_token_address)
            ((reserves_1, reserves_2, _),
             token_address_1, token_address_2, total_supply) = self.ethereum_client.batch_call(
                [
                    pair_contract.functions.getReserves(),
                    pair_contract.functions.token0(),
                    pair_contract.functions.token1(),
                    pair_contract.functions.totalSupply(),
                ])
            decimals_1, decimals_2 = self.get_decimals(token_address_1, token_address_2)

            # Total value for one token should be the same than total value for the other token
            # if pool is under active arbitrage. We use the price for the first token we find
            for token_address, decimals, reserves in zip((token_address_1, token_address_2),
                                                         (decimals_1, decimals_2),
                                                         (reserves_1, reserves_2)):
                try:
                    price = self.get_price(token_address)
                    total_value = (reserves / 10 ** decimals_1) * price
                    return (total_value * 2) / (total_supply / 1e18)
                except CannotGetPriceFromOracle:
                    continue
        except (ValueError, ZeroDivisionError, BadFunctionCallOutput, DecodingError) as e:
            error_message = f'Cannot get uniswap v2 price for pool token={pool_token_address}'
            logger.warning(error_message)
            raise CannotGetPriceFromOracle(error_message) from e


class SushiswapOracle(UniswapV2Oracle):
    pair_init_code = HexBytes('0xe18a34eb0e04b04f7a0ac29a6e80748dca96319b42c54d679cb821dca90c6303')
    router_address = '0xd9e1cE17f2641f24aE83637ab66a2cca9C378B9F'


class AaveOracle(PriceOracle):
    def __init__(self, ethereum_client: EthereumClient, price_oracle: PriceOracle):
        """
        :param ethereum_client:
        :param price_oracle: Price oracle to get the price for the components of the Balancer Pool, UniswapV2 is
        recommended
        """
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self.price_oracle = price_oracle

    def get_price(self, token_address: str) -> float:
        if token_address == '0x4da27a545c0c5B758a6BA100e3a049001de870f5':  # Stacked Aave
            return self.price_oracle.get_price('0x7Fc66500c84A76Ad7e9c93437bFc5Ac33E2DDaE9')
        try:
            underlying_token = self.w3.eth.contract(
                token_address,
                abi=AAVE_ATOKEN_ABI
            ).functions.UNDERLYING_ASSET_ADDRESS().call()
            return self.price_oracle.get_price(underlying_token)
        except (ValueError, BadFunctionCallOutput, DecodingError):
            raise CannotGetPriceFromOracle(f'Cannot get price for {token_address}. It is not an Aaave atoken')


class ZerionComposedOracle(ComposedPriceOracle):
    ZERION_ADAPTER_ADDRESS = None

    def __init__(self, ethereum_client: EthereumClient,
                 zerion_adapter_address: Optional[str] = None):
        """
        :param ethereum_client:
        :param zerion_adapter_address: Can be retrieved using the Zerion registry on
            0x06FE76B2f432fdfEcAEf1a7d4f6C3d41B5861672 . https://github.com/zeriontech/defi-sdk/wiki/Addresses is
            outdated
        """

        self.zerion_adapter_address = zerion_adapter_address or self.ZERION_ADAPTER_ADDRESS
        if not self.zerion_adapter_address:
            raise ValueError('Expected a Zerion adapter address')
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3

    @cached_property
    def zerion_adapter_contract(self) -> Optional[Contract]:
        """
        :return: https://curve.readthedocs.io/registry-registry.html
        """
        if self.ethereum_client.is_contract(self.zerion_adapter_address):
            return self.w3.eth.contract(self.zerion_adapter_address, abi=ZERION_TOKEN_ADAPTER_ABI)

    def get_underlying_tokens(self, token_address: ChecksumAddress) -> List[UnderlyingToken]:
        """
        Use Zerion Token adapter to return underlying components for pool

        :param token_address: Pool token address
        :return: Price per share and underlying token
        :raises: CannotGetPriceFromOracle
        """
        if not self.zerion_adapter_contract:
            raise CannotGetPriceFromOracle(
                f'{self.__class__.__name__}: Cannot get price for {token_address}. Cannot find Zerion adapter'
            )

        try:
            results = self.zerion_adapter_contract.functions.getComponents(token_address).call()
            if results:
                underlying_tokens = []
                for token_address, _, quantity in results:
                    # If there's just one component, quantity must be 100%
                    normalized_quantity = 1. if len(results) == 1 else quantity / 10 ** len(str(quantity))
                    underlying_tokens.append(UnderlyingToken(token_address, normalized_quantity))
                return underlying_tokens
        except (ValueError, BadFunctionCallOutput, DecodingError):
            pass

        raise CannotGetPriceFromOracle(
            f'{self.__class__.__name__}: Cannot get price for {token_address}. It is not a Zerion supported pool token'
        )


class CurveOracle(ZerionComposedOracle):
    """
    Curve pool Oracle. More info on https://curve.fi/
    """
    ZERION_ADAPTER_ADDRESS = '0x99b0bEadc3984eab9842AF81f9fad0C2219108cc'   # Mainnet address


class PoolTogetherOracle(ZerionComposedOracle):
    """
    PoolTogether pool Oracle. More info on https://pooltogether.com/
    """
    ZERION_ADAPTER_ADDRESS = '0xb4E0E1672fFd9b128784dB9f3BE9158fac3f1DFc'   # Mainnet address


class YearnOracle(ComposedPriceOracle):
    """
    Yearn oracle. More info on https://docs.yearn.finance
    """
    def __init__(self, ethereum_client: EthereumClient,
                 yearn_vault_token_adapter: Optional[str] = '0xb460FcC1B6c1CBD7D03F47B6BD5F03994d286c75',
                 iearn_token_adapter: Optional[str] = '0x65B23774daE2a5be02dD275918DDF048d177a5B4'):
        """
        :param ethereum_client:
        :param yearn_vault_token_adapter:
        :param iearn_token_adapter:
        """
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self.yearn_vault_token_adapter = ZerionComposedOracle(ethereum_client, yearn_vault_token_adapter)
        self.iearn_token_adapter = ZerionComposedOracle(ethereum_client, iearn_token_adapter)

    def get_underlying_tokens(self, token_address: ChecksumAddress) -> List[Tuple[float, ChecksumAddress]]:
        """
        :param token_address:
        :return: Price per share and underlying token
        :raises: CannotGetPriceFromOracle
        """
        for adapter in (
            self.yearn_vault_token_adapter,
            self.iearn_token_adapter
        ):
            try:
                # Getting underlying token function is the same for both yVault and yToken
                return adapter.get_underlying_tokens(token_address)
            except CannotGetPriceFromOracle:
                pass

        raise CannotGetPriceFromOracle(f'Cannot get price for {token_address}. It is not a Yearn yToken/yVault')


class BalancerOracle(PricePoolOracle):
    """
    Oracle for Balancer. More info on https://balancer.exchange
    """
    def __init__(self, ethereum_client: EthereumClient, price_oracle: PriceOracle):
        """
        :param ethereum_client:
        :param price_oracle: Price oracle to get the price for the components of the Balancer Pool, UniswapV2 is
        recommended
        """
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self.price_oracle = price_oracle

    def get_pool_token_price(self, pool_token_address: ChecksumAddress) -> float:
        """
        Estimate balancer pool token price based on its components

        :param pool_token_address: Balancer pool token address
        :return: Eth price for pool token
        :raises: CannotGetPriceFromOracle
        """
        try:
            balancer_pool_contract = self.w3.eth.contract(pool_token_address, abi=balancer_pool_abi)
            current_tokens, total_supply = self.ethereum_client.batch_call([
                balancer_pool_contract.functions.getCurrentTokens(),
                balancer_pool_contract.functions.totalSupply(),
            ])
            number_tokens = len(current_tokens)
            # denormalized_weight = self.ethereum_client.batch_call([
            #    balancer_pool_contract.functions.getReserves(current_token)
            #    for current_token in current_tokens
            # ])
            token_balances_and_decimals = self.ethereum_client.batch_call([
                balancer_pool_contract.functions.getBalance(token_address)
                for token_address in current_tokens
            ] + [
                get_erc20_contract(self.w3, token_address).functions.decimals()
                for token_address in current_tokens
            ])
            token_balances = token_balances_and_decimals[:number_tokens]
            token_decimals = token_balances_and_decimals[number_tokens:]
            token_prices = [self.price_oracle.get_price(token_address) for token_address in current_tokens]
            total_eth_value = 0
            for token_balance, token_decimal, token_price in zip(token_balances, token_decimals, token_prices):
                total_eth_value += (token_balance / 10**token_decimal) * token_price
            return total_eth_value / (total_supply / 1e18)
        except (ValueError, BadFunctionCallOutput, DecodingError):
            raise CannotGetPriceFromOracle(f'Cannot get price for {pool_token_address}. '
                                           f'It is not a balancer pool token')


class MooniswapOracle(BalancerOracle):
    def get_pool_token_price(self, pool_token_address: ChecksumAddress) -> float:
        """
        Estimate balancer pool token price based on its components

        :param pool_token_address: Moniswap pool token address
        :return: Eth price for pool token
        :raises: CannotGetPriceFromOracle
        """
        try:
            balancer_pool_contract = self.w3.eth.contract(pool_token_address, abi=mooniswap_abi)
            tokens, total_supply = self.ethereum_client.batch_call([
                balancer_pool_contract.functions.getTokens(),
                balancer_pool_contract.functions.totalSupply(),
            ])
            if len(tokens) == 1 or any([token == NULL_ADDRESS for token in tokens]):  # One of the tokens is ether
                ethereum_amount = self.ethereum_client.get_balance(pool_token_address)
                return ethereum_amount * 2 / total_supply
            else:
                for token in tokens:
                    try:
                        price = self.price_oracle.get_price(token)
                        token_contract = get_erc20_contract(self.w3, token)
                        token_balance, token_decimals = self.ethereum_client.batch_call([
                            token_contract.functions.balanceOf(pool_token_address),
                            token_contract.functions.decimals(),
                        ])
                        total_value = (token_balance / 10 ** token_decimals) * price
                        return (total_value * 2) / (total_supply / 1e18)
                    except CannotGetPriceFromOracle:
                        continue

                raise CannotGetPriceFromOracle(f'Cannot get price for {pool_token_address}. '
                                               f'It is not a mooniswap pool token')

        except (ValueError, BadFunctionCallOutput, DecodingError):
            raise CannotGetPriceFromOracle(f'Cannot get price for {pool_token_address}. '
                                           f'It is not a mooniswap pool token')



from typing import Any, Dict, List, Optional
from urllib.parse import urljoin

import requests
from web3 import Web3

from .. import EthereumNetwork
from .contract_metadata import ContractMetadata


class Sourcify:
    """
    Get contract metadata from Sourcify. Matches can be full or partial:

      - Full: Both the source files as well as the meta data files were an exact match between the deployed bytecode
        and the published files.
      - Partial: Source code compiles to the same bytecode and thus the contract behaves in the same way,
        but the source code can be different: Variables can have misleading names,
        comments can be different and especially the NatSpec comments could have been modified.

    """
    def __init__(self, network: EthereumNetwork = EthereumNetwork.MAINNET,
                 base_url: str = 'https://repo.sourcify.dev/'):
        self.network = network
        self.base_url = base_url
        self.http_session = requests.session()

    def _get_abi_from_metadata(self, metadata: Dict[str, Any]) -> List[Dict[str, Any]]:
        return metadata['output']['abi']

    def _get_name_from_metadata(self, metadata: Dict[str, Any]) -> Optional[str]:
        values = list(metadata['settings'].get('compilationTarget', {}).values())
        if values:
            return values[0]

    def _do_request(self, url: str) -> Optional[Dict[str, Any]]:
        response = self.http_session.get(url, timeout=10)
        if not response.ok:
            return None

        return response.json()

    def get_contract_metadata(self, contract_address: str) -> Optional[ContractMetadata]:
        assert Web3.isChecksumAddress(contract_address), 'Expecting a checksummed address'

        for match_type in ('full_match', 'partial_match'):
            url = urljoin(self.base_url,
                          f'/contracts/{match_type}/{self.network.value}/{contract_address}/metadata.json')
            metadata = self._do_request(url)
            if metadata:
                abi = self._get_abi_from_metadata(metadata)
                name = self._get_name_from_metadata(metadata)
                return ContractMetadata(name, abi, match_type == 'partial_match')
        return None



from django.test import TestCase

from eth_abi.packed import encode_abi_packed
from eth_account import Account
from hexbytes import HexBytes

from ..contracts import (get_proxy_1_0_0_deployed_bytecode,
                         get_proxy_factory_contract)
from ..utils import (compare_byte_code, decode_string_or_bytes32,
                     generate_address_2)
from .ethereum_test_case import EthereumTestCaseMixin


class TestUtils(EthereumTestCaseMixin, TestCase):
    def test_generate_address_2(self):
        from_ = '0x8942595A2dC5181Df0465AF0D7be08c8f23C93af'
        salt = self.w3.keccak(text='aloha')
        init_code = '0x00abcd'
        expected = '0x8D02C796Dd019916F65EBa1C9D65a7079Ece00E0'
        address2 = generate_address_2(from_, salt, init_code)
        self.assertEqual(address2, expected)

        from_ = HexBytes('0x8942595A2dC5181Df0465AF0D7be08c8f23C93af')
        salt = self.w3.keccak(text='aloha').hex()
        init_code = HexBytes('0x00abcd')
        expected = '0x8D02C796Dd019916F65EBa1C9D65a7079Ece00E0'
        address2 = generate_address_2(from_, salt, init_code)
        self.assertEqual(address2, expected)

    def test_generate_address2_with_proxy(self):
        deployer_account = self.ethereum_test_account
        proxy_factory_contract = get_proxy_factory_contract(self.w3)
        nonce = self.w3.eth.get_transaction_count(deployer_account.address, block_identifier='pending')
        tx = proxy_factory_contract.constructor().buildTransaction({'nonce': nonce,
                                                                    'from': deployer_account.address})
        signed_tx = deployer_account.sign_transaction(tx)
        tx_hash = self.w3.eth.send_raw_transaction(signed_tx.rawTransaction)
        tx_receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash)
        proxy_factory_contract = get_proxy_factory_contract(self.w3, address=tx_receipt['contractAddress'])

        initializer = b''  # Should be the safe `setup()` call with `owners`, `threshold`, `payment`...
        salt_nonce = 0  # Random. For sure. I used a dice

        master_copy = Account.create().address
        tx = proxy_factory_contract.functions.createProxyWithNonce(master_copy,
                                                                   initializer,
                                                                   salt_nonce
                                                                   ).buildTransaction({'nonce': nonce + 1,
                                                                                       'from': deployer_account.address,
                                                                                       })
        signed_tx = deployer_account.sign_transaction(tx)
        tx_hash = self.w3.eth.send_raw_transaction(signed_tx.rawTransaction)
        tx_receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash)
        logs = proxy_factory_contract.events.ProxyCreation().processReceipt(tx_receipt)
        log = logs[0]
        self.assertEqual(log['event'], 'ProxyCreation')
        proxy_address = log['args']['proxy']

        proxy_creation_code = proxy_factory_contract.functions.proxyCreationCode().call()
        salt = self.w3.keccak(encode_abi_packed(['bytes', 'uint256'], [self.w3.keccak(initializer), salt_nonce]))
        deployment_data = encode_abi_packed(['bytes', 'uint256'], [proxy_creation_code, int(master_copy, 16)])
        address2 = generate_address_2(proxy_factory_contract.address, salt, deployment_data)
        self.assertEqual(proxy_address, address2)

    def test_decode_string_or_bytes32(self):
        # Abi encoded string
        gnosis_hex = HexBytes('0x00000000000000000000000000000000000000000000000000000000000000200000000000000000000000'
                              '00000000000000000000000000000000000000000c476e6f73697320546f6b656e0000000000000000000000'
                              '000000000000000000')

        # Abi encoded bytes32
        dai_hex = HexBytes('0x44616920537461626c65636f696e2076312e3000000000000000000000000000')

        self.assertEqual(decode_string_or_bytes32(gnosis_hex), 'Gnosis Token')
        self.assertEqual(decode_string_or_bytes32(dai_hex), 'Dai Stablecoin v1.0')

    def test_compare_byte_code(self):
        proxy_with_metadata = get_proxy_1_0_0_deployed_bytecode()
        proxy_with_different_metadata = HexBytes('0x608060405273ffffffffffffffffffffffffffffffffffffffff60005416366000'
                                                 '8037600080366000845af43d6000803e6000811415603d573d6000fd5b3d6000f3fe'
                                                 'a165627a7a72305820b7f4e514a2bdfeb2e729e84f7101233a43b51d677007041e70'
                                                 '8067e4b88bec480029')
        self.assertTrue(compare_byte_code(proxy_with_metadata, proxy_with_different_metadata))



from ethereum.transactions import secpk1n

NULL_ADDRESS: str = '0x' + '0' * 40
SENTINEL_ADDRESS: str = '0x' + '0' * 39 + '1'

# keccak('Transfer(address,address,uint256)')
ERC20_721_TRANSFER_TOPIC: str = '0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'

SIGNATURE_R_MIN_VALUE: int = 1
SIGNATURE_R_MAX_VALUE: int = secpk1n - 1
SIGNATURE_S_MIN_VALUE: int = 1
SIGNATURE_S_MAX_VALUE: int = secpk1n // 2
SIGNATURE_V_MIN_VALUE: int = 27
SIGNATURE_V_MAX_VALUE: int = 28


GAS_CALL_DATA_ZERO_BYTE = 4
GAS_CALL_DATA_BYTE = 16  # 68 before Istanbul



from django.conf import settings
from django.test import TestCase

from eth_account import Account

from gnosis.eth import EthereumClient, InsufficientFunds


class TestRealNodes(TestCase):
    def test_node_exceptions(self):
        node_urls = settings.ETHEREUM_TEST_NODES_URLS
        for node_url in node_urls:
            random_address = Account.create().address
            random_sender_account = Account.create()
            ethereum_service = EthereumClient(node_url)
            with self.assertRaises(InsufficientFunds):
                ethereum_service.send_unsigned_transaction({'to': random_address, 'value': 0, 'data': b'',
                                                            'gas': 25000, 'gasPrice': 1},
                                                           private_key=random_sender_account.key)

    def test_safe_exceptions(self):
        pass



# flake8: noqa F401
from .exceptions import (CannotEstimateGas, CouldNotPayGasWithEther,
                         InvalidChecksumAddress, InvalidInternalTx,
                         InvalidMultisigTx, InvalidPaymentToken,
                         InvalidSignaturesProvided, SafeServiceException,
                         SignatureNotProvidedByOwner)
from .proxy_factory import ProxyFactory
from .safe import Safe, SafeOperation, SafeTx



import math
import os
from logging import getLogger
from typing import Any, Dict, List, Optional, Tuple

import rlp
from ethereum.exceptions import InvalidTransaction
from ethereum.transactions import Transaction, secpk1n
from ethereum.utils import checksum_encode, mk_contract_address
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import ContractConstructor

from gnosis.eth.constants import GAS_CALL_DATA_BYTE, NULL_ADDRESS
from gnosis.eth.contracts import (get_erc20_contract,
                                  get_paying_proxy_contract,
                                  get_safe_V0_0_1_contract)

logger = getLogger(__name__)


class InvalidERC20Token(Exception):
    pass


class SafeCreationTx:
    def __init__(self, w3: Web3, owners: List[str], threshold: int, signature_s: int, master_copy: str,
                 gas_price: int, funder: Optional[str], payment_token: Optional[str] = None,
                 payment_token_eth_value: float = 1.0, fixed_creation_cost: Optional[int] = None):
        """
        Prepare Safe creation
        :param w3: Web3 instance
        :param owners: Owners of the Safe
        :param threshold: Minimum number of users required to operate the Safe
        :param signature_s: Random s value for ecdsa signature
        :param master_copy: Safe master copy address
        :param gas_price: Gas Price
        :param funder: Address to refund when the Safe is created. Address(0) if no need to refund
        :param payment_token: Payment token instead of paying the funder with ether. If None Ether will be used
        :param payment_token_eth_value: Value of payment token per 1 Ether
        :param fixed_creation_cost: Fixed creation cost of Safe (Wei)
        """

        assert 0 < threshold <= len(owners)
        funder = funder or NULL_ADDRESS
        payment_token = payment_token or NULL_ADDRESS
        assert Web3.isChecksumAddress(master_copy)
        assert Web3.isChecksumAddress(funder)
        assert Web3.isChecksumAddress(payment_token)

        self.w3 = w3
        self.owners = owners
        self.threshold = threshold
        self.s = signature_s
        self.master_copy = master_copy
        self.gas_price = gas_price
        self.funder = funder
        self.payment_token = payment_token
        self.payment_token_eth_value = payment_token_eth_value
        self.fixed_creation_cost = fixed_creation_cost

        # Get bytes for `setup(address[] calldata _owners, uint256 _threshold, address to, bytes calldata data)`
        # This initializer will be passed to the proxy and will be called right after proxy is deployed
        safe_setup_data: bytes = self._get_initial_setup_safe_data(owners, threshold)

        # Calculate gas based on experience of previous deployments of the safe
        calculated_gas: int = self._calculate_gas(owners, safe_setup_data, payment_token)
        # Estimate gas using web3
        estimated_gas: int = self._estimate_gas(master_copy, safe_setup_data, funder, payment_token)
        self.gas = max(calculated_gas, estimated_gas)

        # Payment will be safe deploy cost + transfer fees for sending ether to the deployer
        self.payment = self._calculate_refund_payment(self.gas,
                                                      gas_price,
                                                      fixed_creation_cost,
                                                      payment_token_eth_value)

        self.tx_dict: Dict[str, Any] = self._build_proxy_contract_creation_tx(master_copy=master_copy,
                                                                              initializer=safe_setup_data,
                                                                              funder=funder,
                                                                              payment_token=payment_token,
                                                                              payment=self.payment,
                                                                              gas=self.gas,
                                                                              gas_price=gas_price)

        self.tx_pyethereum: Transaction = self._build_contract_creation_tx_with_valid_signature(self.tx_dict, self.s)
        self.tx_raw = rlp.encode(self.tx_pyethereum)
        self.tx_hash = self.tx_pyethereum.hash
        self.deployer_address = checksum_encode(self.tx_pyethereum.sender)
        self.safe_address = checksum_encode(self.tx_pyethereum.creates)

        self.v = self.tx_pyethereum.v
        self.r = self.tx_pyethereum.r
        self.safe_setup_data = safe_setup_data

        assert checksum_encode(mk_contract_address(self.deployer_address, nonce=0)) == self.safe_address

    @property
    def payment_ether(self):
        return self.gas * self.gas_price

    @staticmethod
    def find_valid_random_signature(s: int) -> Tuple[int, int]:
        """
        Find v and r valid values for a given s
        :param s: random value
        :return: v, r
        """
        for _ in range(10000):
            r = int(os.urandom(31).hex(), 16)
            v = (r % 2) + 27
            if r < secpk1n:
                tx = Transaction(0, 1, 21000, b'', 0, b'', v=v, r=r, s=s)
                try:
                    tx.sender
                    return v, r
                except (InvalidTransaction, ValueError):
                    logger.debug('Cannot find signature with v=%d r=%d s=%d', v, r, s)

        raise ValueError('Valid signature not found with s=%d', s)

    @staticmethod
    def _calculate_gas(owners: List[str], safe_setup_data: bytes, payment_token: str) -> int:
        """
        Calculate gas manually, based on tests of previosly deployed safes
        :param owners: Safe owners
        :param safe_setup_data: Data for proxy setup
        :param payment_token: If payment token, we will need more gas to transfer and maybe storage if first time
        :return: total gas needed for deployment
        """
        # TODO Do gas calculation estimating the call instead this magic

        base_gas = 60580  # Transaction standard gas

        # If we already have the token, we don't have to pay for storage, so it will be just 5K instead of 20K.
        # The other 1K is for overhead of making the call
        if payment_token != NULL_ADDRESS:
            payment_token_gas = 55000
        else:
            payment_token_gas = 0

        data_gas = GAS_CALL_DATA_BYTE * len(safe_setup_data)  # Data gas
        gas_per_owner = 18020  # Magic number calculated by testing and averaging owners
        return base_gas + data_gas + payment_token_gas + 270000 + len(owners) * gas_per_owner

    @staticmethod
    def _calculate_refund_payment(gas: int, gas_price: int, fixed_creation_cost: Optional[int],
                                  payment_token_eth_value: float) -> int:
        if fixed_creation_cost is None:
            # Payment will be safe deploy cost + transfer fees for sending ether to the deployer
            base_payment: int = (gas + 23000) * gas_price
            # Calculate payment for tokens using the conversion (if used)
            return math.ceil(base_payment / payment_token_eth_value)
        else:
            return fixed_creation_cost

    def _build_proxy_contract_creation_constructor(self,
                                                   master_copy: str,
                                                   initializer: bytes,
                                                   funder: str,
                                                   payment_token: str,
                                                   payment: int) -> ContractConstructor:
        """
        :param master_copy: Master Copy of Gnosis Safe already deployed
        :param initializer: Data initializer to send to GnosisSafe setup method
        :param funder: Address that should get the payment (if payment set)
        :param payment_token: Address if a token is used. If not set, 0x0 will be ether
        :param payment: Payment
        :return: Transaction dictionary
        """
        if not funder or funder == NULL_ADDRESS:
            funder = NULL_ADDRESS
            payment = 0

        return get_paying_proxy_contract(self.w3).constructor(
            master_copy,
            initializer,
            funder,
            payment_token,
            payment)

    def _build_proxy_contract_creation_tx(self,
                                          master_copy: str,
                                          initializer: bytes,
                                          funder: str,
                                          payment_token: str,
                                          payment: int,
                                          gas: int,
                                          gas_price: int,
                                          nonce: int = 0):
        """
        :param master_copy: Master Copy of Gnosis Safe already deployed
        :param initializer: Data initializer to send to GnosisSafe setup method
        :param funder: Address that should get the payment (if payment set)
        :param payment_token: Address if a token is used. If not set, 0x0 will be ether
        :param payment: Payment
        :return: Transaction dictionary
        """
        return self._build_proxy_contract_creation_constructor(
            master_copy,
            initializer,
            funder,
            payment_token,
            payment
        ).buildTransaction({
            'gas': gas,
            'gasPrice': gas_price,
            'nonce': nonce,
        })

    def _build_contract_creation_tx_with_valid_signature(self, tx_dict: Dict[str, Any], s: int) -> Transaction:
        """
        Use pyethereum `Transaction` to generate valid tx using a random signature
        :param tx_dict: Web3 tx dictionary
        :param s: Signature s value
        :return: PyEthereum creation tx for the proxy contract
        """
        zero_address = HexBytes('0x' + '0' * 40)
        f_address = HexBytes('0x' + 'f' * 40)
        nonce = tx_dict['nonce']
        gas_price = tx_dict['gasPrice']
        gas = tx_dict['gas']
        to = tx_dict.get('to', b'')  # Contract creation should always have `to` empty
        value = tx_dict['value']
        data = tx_dict['data']
        for _ in range(100):
            try:
                v, r = self.find_valid_random_signature(s)
                contract_creation_tx = Transaction(nonce, gas_price, gas, to, value, HexBytes(data), v=v, r=r, s=s)
                sender_address = contract_creation_tx.sender
                contract_address = contract_creation_tx.creates
                if sender_address in (zero_address, f_address) or contract_address in (zero_address, f_address):
                    raise InvalidTransaction
                return contract_creation_tx
            except InvalidTransaction:
                pass
        raise ValueError('Valid signature not found with s=%d', s)

    def _estimate_gas(self,
                      master_copy: str,
                      initializer: bytes,
                      funder: str,
                      payment_token: str) -> int:
        """
        Gas estimation done using web3 and calling the node
        Payment cannot be estimated, as no ether is in the address. So we add some gas later.
        :param master_copy: Master Copy of Gnosis Safe already deployed
        :param initializer: Data initializer to send to GnosisSafe setup method
        :param funder: Address that should get the payment (if payment set)
        :param payment_token: Address if a token is used. If not set, 0x0 will be ether
        :return: Total gas estimation
        """

        # Estimate the contract deployment. We cannot estimate the refunding, as the safe address has not any fund
        gas: int = self._build_proxy_contract_creation_constructor(
            master_copy,
            initializer,
            funder,
            payment_token,
            0).estimateGas()

        # We estimate the refund as a new tx
        if payment_token == NULL_ADDRESS:
            # Same cost to send 1 ether than 1000
            gas += self.w3.eth.estimate_gas({'to': funder, 'value': 1})
        else:
            # Top should be around 52000 when storage is needed (funder no previous owner of token),
            # we use value 1 as we are simulating an internal call, and in that calls you don't pay for the data.
            # If it was a new tx sending 5000 tokens would be more expensive than sending 1 because of data costs
            try:
                gas += get_erc20_contract(self.w3,
                                          payment_token).functions.transfer(funder, 1).estimateGas({'from':
                                                                                                    payment_token})
            except ValueError as exc:
                if 'transfer amount exceeds balance' in str(exc):
                    return 70000
                raise InvalidERC20Token from exc

        return gas

    def _get_initial_setup_safe_data(self, owners: List[str], threshold: int) -> bytes:
        return get_safe_V0_0_1_contract(self.w3, self.master_copy).functions.setup(
            owners,
            threshold,
            NULL_ADDRESS,  # Contract address for optional delegate call
            b''            # Data payload for optional delegate call
        ).buildTransaction({
            'gas': 1,
            'gasPrice': 1,
        })['data']



from django.test import TestCase

from ethereum.utils import check_checksum, sha3
from faker import Faker
from hexbytes import HexBytes

from ...utils import get_eth_address_with_key
from .models import EthereumAddress, Sha3Hash, Uint256

faker = Faker()


class TestModels(TestCase):
    def test_ethereum_address_field(self):
        address, _ = get_eth_address_with_key()
        self.assertTrue(check_checksum(address))
        ethereum_address = EthereumAddress.objects.create(value=address)
        ethereum_address.refresh_from_db()
        self.assertTrue(check_checksum(ethereum_address.value))
        self.assertEqual(address, ethereum_address.value)

        ethereum_address = EthereumAddress.objects.create(value=None)
        ethereum_address.refresh_from_db()
        self.assertIsNone(ethereum_address.value)

        with self.assertRaises(Exception):
            EthereumAddress.objects.create(value='0x23')

    def test_uint256_field(self):
        for value in [2, -2, 2 ** 256, 2 ** 260,
                      25572735541615049941137326092682691158109824779649981270427004917341670006487,
                      None]:
            uint256 = Uint256.objects.create(value=value)
            uint256.refresh_from_db()
            self.assertEqual(uint256.value, value)

        # Overflow
        with self.assertRaises(Exception):
            value = 2 ** 263
            Uint256.objects.create(value=value)

    def test_sha3_hash_field(self):
        value: bytes = sha3(faker.name())
        value_hex_without_0x: str = value.hex()
        value_hex_with_0x: str = '0x' + value_hex_without_0x
        value_hexbytes: HexBytes = HexBytes(value_hex_with_0x)

        values = [value, value_hex_without_0x, value_hex_with_0x, value_hexbytes]

        for v in values:
            sha3_hash = Sha3Hash.objects.create(value=v)
            sha3_hash.refresh_from_db()
            self.assertEqual(sha3_hash.value, value_hex_with_0x)

        for v in values:
            self.assertEqual(Sha3Hash.objects.filter(value=v).count(), len(values))

        # Hash null
        sha3_hash = Sha3Hash.objects.create(value=None)
        sha3_hash.refresh_from_db()
        self.assertIsNone(sha3_hash.value)

        # Hash too big
        value_hex_invalid: str = '0x' + value_hex_without_0x + 'a'
        with self.assertRaises(Exception):
            Sha3Hash.objects.create(value=value_hex_invalid)



import logging

from django.test import TestCase

from eth_account import Account
from hexbytes import HexBytes

from ...eth.contracts import get_safe_contract
from ..exceptions import NotEnoughSafeTransactionGas, SignaturesDataTooShort
from ..multi_send import MultiSendOperation, MultiSendTx
from ..safe import Safe, SafeOperation
from ..safe_tx import SafeTx
from .safe_test_case import SafeTestCaseMixin

logger = logging.getLogger(__name__)


class TestSafeTx(SafeTestCaseMixin, TestCase):
    def test_multi_send_safe_tx(self):
        owners = [Account.create() for _ in range(2)]
        owner_addresses = [owner.address for owner in owners]
        threshold = 1
        safe_creation = self.deploy_test_safe(owners=owner_addresses, threshold=threshold,
                                              initial_funding_wei=self.w3.toWei(0.1, 'ether'))
        safe_address = safe_creation.safe_address
        safe = Safe(safe_address, self.ethereum_client)
        safe_contract = safe.get_contract()
        to = self.multi_send_contract.address
        value = 0
        safe_tx_gas = 600000
        base_gas = 200000

        # Atomic swap the owner of a Safe
        new_owner = Account.create()
        owner_to_remove = owners[-1]
        prev_owner = owners[-2]
        owners_expected = [x.address for x in owners[:-1]] + [new_owner.address]
        new_threshold = threshold + 1
        data = HexBytes(safe_contract.functions.addOwnerWithThreshold(new_owner.address,
                                                                      new_threshold).buildTransaction({'gas': 0})['data'])
        data_2 = HexBytes(safe_contract.functions.removeOwner(prev_owner.address, owner_to_remove.address,
                                                              new_threshold).buildTransaction({'gas': 0})['data'])

        multisend_txs = [MultiSendTx(MultiSendOperation.CALL, safe_address, value, d) for d in (data, data_2)]
        safe_multisend_data = self.multi_send.build_tx_data(multisend_txs)
        safe_tx = SafeTx(self.ethereum_client, safe_address, to,
                         0, safe_multisend_data, SafeOperation.DELEGATE_CALL.value,
                         safe_tx_gas, base_gas, self.gas_price, None, None, safe_nonce=0)
        safe_tx.sign(owners[0].key)

        self.assertEqual(safe_tx.call(tx_sender_address=self.ethereum_test_account.address), 1)
        tx_hash, _ = safe_tx.execute(tx_sender_private_key=self.ethereum_test_account.key)
        self.ethereum_client.get_transaction_receipt(tx_hash, timeout=60)
        self.assertEqual(safe.retrieve_nonce(), 1)
        self.assertEqual(safe.retrieve_threshold(), new_threshold)
        self.assertCountEqual(safe.retrieve_owners(), owners_expected)

    def test_send_safe_tx(self):
        owners = [Account.create() for _ in range(2)]
        owner_addresses = [owner.address for owner in owners]
        threshold = 1
        safe_creation = self.deploy_test_safe(owners=owner_addresses, threshold=threshold,
                                              initial_funding_wei=self.w3.toWei(0.1, 'ether'))
        safe_address = safe_creation.safe_address
        to = Account().create().address
        value = self.w3.toWei(0.01, 'ether')
        safe_tx_gas = 200000
        data_gas = 100000

        safe_tx = SafeTx(self.ethereum_client, safe_address, to, value, b'', 0, safe_tx_gas, data_gas, self.gas_price,
                         None, None, safe_nonce=0)

        with self.assertRaises(SignaturesDataTooShort):
            safe_tx.call(tx_sender_address=self.ethereum_test_account.address)

        # Check signing
        self.assertFalse(safe_tx.signers)
        safe_tx.sign(owners[0].key)
        self.assertIn(owners[0].address, safe_tx.signers)

        with self.assertRaises(NotEnoughSafeTransactionGas):
            safe_tx.call(tx_sender_address=self.ethereum_test_account.address, tx_gas=safe_tx_gas // 2)

        self.assertEqual(safe_tx.call(tx_sender_address=self.ethereum_test_account.address), 1)
        tx_hash, _ = safe_tx.execute(tx_sender_private_key=self.ethereum_test_account.key)
        self.ethereum_client.get_transaction_receipt(tx_hash, timeout=60)
        self.assertEqual(self.ethereum_client.get_balance(to), value)

        safe_tx.unsign(owners[0].address)
        self.assertFalse(safe_tx.signers)

    def test_sign_safe_tx(self):
        owners = [Account.create() for _ in range(3)]
        owners_unsorted = sorted(owners, key=lambda x: int(x.address, 16), reverse=True)
        owner_addresses = [owner.address for owner in owners_unsorted]
        threshold = 1
        safe_creation = self.deploy_test_safe(owners=owner_addresses, threshold=threshold,
                                              initial_funding_wei=self.w3.toWei(0.1, 'ether'))
        safe_address = safe_creation.safe_address
        to = Account().create().address
        value = self.w3.toWei(0.01, 'ether')

        safe_tx = SafeTx(self.ethereum_client, safe_address, to, value, b'', 0, 200000, 100000, self.gas_price,
                         None, None, safe_nonce=0)

        safe_tx.sign(owners_unsorted[0].key)
        safe_tx.sign(owners_unsorted[2].key)
        signers = [owner_addresses[0], owner_addresses[2]]

        self.assertEqual(safe_tx.signers, safe_tx.sorted_signers)
        self.assertNotEqual(signers, safe_tx.signers)
        self.assertEqual(set(signers), set(safe_tx.signers))
        self.assertEqual(len(safe_tx.signers), 2)

        safe_tx.sign(owners_unsorted[1].key)
        signers = owner_addresses
        self.assertEqual(safe_tx.signers, safe_tx.sorted_signers)
        self.assertNotEqual(signers, safe_tx.signers)
        self.assertEqual(set(signers), set(safe_tx.signers))
        self.assertEqual(len(safe_tx.signers), 3)

        # Sign again
        safe_tx.sign(owners_unsorted[0].key)
        self.assertEqual(len(safe_tx.signers), 3)

        # Sign again
        safe_tx.unsign(owners_unsorted[1].address)
        signers = [owner_addresses[0], owner_addresses[2]]
        self.assertEqual(set(signers), set(safe_tx.signers))
        self.assertEqual(len(safe_tx.signers), 2)

    def test_hash_safe_multisig_tx(self):
        # -------- Old version of the contract --------------------------
        expected_hash = HexBytes('0xc9d69a2350aede7978fdee58e702647e4bbdc82168577aa4a43b66ad815c6d1a')
        tx_hash = SafeTx(self.ethereum_client, '0x692a70d2e424a56d2c6c27aa97d1a86395877b3a',
                         '0x5AC255889882aaB35A2aa939679E3F3d4Cea221E',
                         5000000,
                         HexBytes('0x00'),
                         0,
                         50000,
                         100,
                         10000,
                         '0x' + '0' * 40,
                         '0x' + '0' * 40,
                         safe_nonce=67, safe_version='0.1.0').safe_tx_hash
        self.assertEqual(expected_hash, tx_hash)

        expected_hash = HexBytes('0x8ca8db91d72b379193f6e229eb2dff0d0621b6ef452d90638ee3206e9b7349b3')
        tx_hash = SafeTx(self.ethereum_client, '0x692a70d2e424a56d2c6c27aa97d1a86395877b3a',
                         '0x' + '0' * 40,
                         80000000,
                         HexBytes('0x562944'),
                         2,
                         54522,
                         773,
                         22000000,
                         '0x' + '0' * 40,
                         '0x' + '0' * 40,
                         safe_nonce=257000, safe_version='0.1.0').safe_tx_hash
        self.assertEqual(expected_hash, tx_hash)

        # -------- New version of the contract --------------------------
        expected_hash = HexBytes('0x7c60341f3e1b4483575f38e84e97d6b332a2dd55b9290f39e6e26eef29a04fe7')
        tx_hash = SafeTx(self.ethereum_client, '0x692a70D2e424a56D2C6C27aA97D1a86395877b3A',
                         '0x5AC255889882aaB35A2aa939679E3F3d4Cea221E',
                         5000000,
                         HexBytes('0x00'),
                         0,
                         50000,
                         100,
                         10000,
                         '0x' + '0' * 40,
                         '0x' + '0' * 40,
                         safe_nonce=67,
                         safe_version='1.0.0').safe_tx_hash
        self.assertEqual(expected_hash, tx_hash)

        expected_hash = HexBytes('0xf585279fd867c94738096f4eab964e9e202014d2f0d5155d751099ad85cbe504')
        tx_hash = SafeTx(self.ethereum_client, '0x692a70D2e424a56D2C6C27aA97D1a86395877b3A',
                         '0x' + '0' * 40,
                         80000000,
                         HexBytes('0x562944'),
                         2,
                         54522,
                         773,
                         22000000,
                         '0x' + '0' * 40,
                         '0x' + '0' * 40,
                         safe_nonce=257000,
                         safe_version='1.0.0').safe_tx_hash
        self.assertEqual(expected_hash, tx_hash)

        safe_create2_tx = self.deploy_test_safe()
        safe_address = safe_create2_tx.safe_address
        # Expected hash must be the same calculated by `getTransactionHash` of the contract
        expected_hash = get_safe_contract(self.ethereum_client.w3, safe_address).functions.getTransactionHash(
            '0x5AC255889882aaB35A2aa939679E3F3d4Cea221E',
            5212459,
            HexBytes(0x00),
            1,
            123456,
            122,
            12345,
            '0x' + '2' * 40,
            '0x' + '2' * 40,
            10789).call()
        safe_tx_hash = SafeTx(self.ethereum_client, safe_address,
                              '0x5AC255889882aaB35A2aa939679E3F3d4Cea221E',
                              5212459,
                              HexBytes(0x00),
                              1,
                              123456,
                              122,
                              12345,
                              '0x' + '2' * 40,
                              '0x' + '2' * 40,
                              safe_nonce=10789,
                              safe_version='1.0.0').safe_tx_hash
        self.assertEqual(HexBytes(expected_hash), safe_tx_hash)

    def test_hash_safe_multisig_tx_1_3(self):
        safe_address = '0x2B0b9cBDA3D7b0F760187c52A8FFB18C48E5d96A'
        expected_hash = '0x01d42a801ac44d3ebd43592be980dea0756d7f7b27d718d3016a42ea7ce85587'
        to = '0x79613FD49472C3C7a32188e45ff00e7bdC8a897d'
        data = HexBytes('0x1212')
        value = 2
        operation = 0
        safe_tx_gas = 0
        base_gas = 0
        gas_price = 0
        gas_token = '0xFA995c7a0d32A4e7497508a5C380369BE8dB49Db'
        refund_receiver = '0x6b92f5E5360bfCa8F5d3FcE65D87382967847983'
        safe_nonce = 17
        chain_id = 4  # Rinkeby
        safe_tx_hash = SafeTx(self.ethereum_client, safe_address,
                              to,
                              value,
                              data,
                              operation,
                              safe_tx_gas,
                              base_gas,
                              gas_price,
                              gas_token,
                              refund_receiver,
                              safe_nonce=safe_nonce,
                              safe_version='1.3.0',
                              chain_id=chain_id,
                              ).safe_tx_hash
        self.assertEqual(HexBytes(expected_hash), safe_tx_hash)

        # Test with ganache master copy contract v1.3.0
        chain_id = self.ethereum_client.w3.eth.chain_id
        safe_tx_hash = SafeTx(self.ethereum_client, self.safe_contract_V1_3_0_address,
                              to,
                              value,
                              data,
                              operation,
                              safe_tx_gas,
                              base_gas,
                              gas_price,
                              gas_token,
                              refund_receiver,
                              safe_nonce=safe_nonce,
                              safe_version='1.3.0',
                              chain_id=chain_id,
                              ).safe_tx_hash
        expected_hash = self.safe_contract_V1_3_0.functions.getTransactionHash(
            to,
            value,
            data,
            operation,
            safe_tx_gas,
            base_gas,
            gas_price,
            gas_token,
            refund_receiver,
            safe_nonce
        ).call()
        self.assertEqual(HexBytes(expected_hash), safe_tx_hash)



from typing import Any, Dict, List, Sequence
from unittest import mock
from unittest.mock import MagicMock

from django.test import TestCase

from eth_account import Account
from hexbytes import HexBytes
from web3.datastructures import AttributeDict
from web3.net import Net

from ..constants import GAS_CALL_DATA_BYTE
from ..contracts import get_erc20_contract
from ..ethereum_client import (EthereumClientProvider, EthereumNetwork,
                               FromAddressNotFound, InsufficientFunds,
                               InvalidERC20Info, InvalidNonce, ParityManager,
                               SenderAccountNotFoundInNode)
from ..utils import get_eth_address_with_key
from .ethereum_test_case import EthereumTestCaseMixin
from .mocks.mock_internal_txs import (creation_internal_txs,
                                      internal_txs_errored)


class TestERC20Module(EthereumTestCaseMixin, TestCase):
    def test_decode_transfer_log(self):
        logs = [AttributeDict({'address': '0x39C4BFa00b6edecCDd00fA9589E1BE76DE63e862',
                               'topics': [HexBytes('0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'),
                                          HexBytes('0x00000000000000000000000094e01661ebaef430fe862f958c03200b0f483f27'),
                                          HexBytes('0x00000000000000000000000094e01661ebaef430fe862f958c03200b0f483f27')],
                               'data': '0x000000000000000000000000000000000000000000000003aa2371d700680000',
                               'blockNumber': 4357126,
                               'transactionHash': HexBytes('0x21381484d8f69dcd782560d1fd3cd818e743c79767985d01aec7e61c2a7f1de9'),
                               'transactionIndex': 14,
                               'blockHash': HexBytes('0x677ada1a306fc50751001bca6eeaa3f5a87a0bf2c9f6fa27899bfbaf999cca4f'),
                               'logIndex': 13,
                               'removed': False}),
                AttributeDict({'address': '0x39C4BFa00b6edecCDd00fA9589E1BE76DE63e862',
                               'topics': [HexBytes('0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'),
                                          HexBytes('0x00000000000000000000000094e01661ebaef430fe862f958c03200b0f483f27'),
                                          HexBytes('0x00000000000000000000000064da772dd84965f0ee58174941d78a9dfbccca2e')],
                               'data': '0x000000000000000000000000000000000000000000000001a055690d9db80000',
                               'blockNumber': 4357126,
                               'transactionHash': HexBytes('0x21381484d8f69dcd782560d1fd3cd818e743c79767985d01aec7e61c2a7f1de9'),
                               'transactionIndex': 14,
                               'blockHash': HexBytes('0x677ada1a306fc50751001bca6eeaa3f5a87a0bf2c9f6fa27899bfbaf999cca4f'),
                               'logIndex': 14,
                               'removed': False}),
                AttributeDict({'address': '0x39C4BFa00b6edecCDd00fA9589E1BE76DE63e862',
                               'topics': [HexBytes('0x8c5be1e5ebec7d5bd14f71427d1e84f3dd0314c0f7b2291e5b200ac8c7c3b925'),
                                          HexBytes('0x00000000000000000000000064da772dd84965f0ee58174941d78a9dfbccca2e'),
                                          HexBytes('0x00000000000000000000000080cdad25de6b439e866805b2dc6808d23ff57b5d')],
                               'data': '0x000000000000000000000000000000000000000000000001a055690d9db80000',
                               'blockNumber': 4357126,
                               'transactionHash': HexBytes('0x21381484d8f69dcd782560d1fd3cd818e743c79767985d01aec7e61c2a7f1de9'),
                               'transactionIndex': 14,
                               'blockHash': HexBytes('0x677ada1a306fc50751001bca6eeaa3f5a87a0bf2c9f6fa27899bfbaf999cca4f'),
                               'logIndex': 15,
                               'removed': False}),
                AttributeDict({'address': '0x39C4BFa00b6edecCDd00fA9589E1BE76DE63e862',
                               'topics': [HexBytes('0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'),
                                          HexBytes('0x00000000000000000000000064da772dd84965f0ee58174941d78a9dfbccca2e'),
                                          HexBytes('0x00000000000000000000000080cdad25de6b439e866805b2dc6808d23ff57b5d')],
                               'data': '0x000000000000000000000000000000000000000000000001a055690d9db80000',
                               'blockNumber': 4357126,
                               'transactionHash': HexBytes('0x21381484d8f69dcd782560d1fd3cd818e743c79767985d01aec7e61c2a7f1de9'),
                               'transactionIndex': 14,
                               'blockHash': HexBytes('0x677ada1a306fc50751001bca6eeaa3f5a87a0bf2c9f6fa27899bfbaf999cca4f'),
                               'logIndex': 16,
                               'removed': False}),
                AttributeDict({'address': '0x99b9F9BA62002a9b43aF6e540428277D5E52EF47',
                               'topics': [HexBytes('0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'),
                                          HexBytes('0x00000000000000000000000099b9f9ba62002a9b43af6e540428277d5e52ef47'),
                                          HexBytes('0x00000000000000000000000094e01661ebaef430fe862f958c03200b0f483f27'),
                                          HexBytes('0xcc292d3dab2c0fbbf616670ac57ec51162959c2d9cbe938819b6e8bc1c757335')],
                               'data': '0x',
                               'blockNumber': 4357126,
                               'transactionHash': HexBytes('0x21381484d8f69dcd782560d1fd3cd818e743c79767985d01aec7e61c2a7f1de9'),
                               'transactionIndex': 14,
                               'blockHash': HexBytes('0x677ada1a306fc50751001bca6eeaa3f5a87a0bf2c9f6fa27899bfbaf999cca4f'),
                               'logIndex': 17,
                               'removed': False}),
                AttributeDict({'address': '0x99b9F9BA62002a9b43aF6e540428277D5E52EF47',
                               'topics': [HexBytes('0x2114851a3e2a54429989f46c1ab0743e37ded205d9bbdfd85635aed5bd595a06'),
                                          HexBytes('0x00000000000000000000000099b9f9ba62002a9b43af6e540428277d5e52ef47'),
                                          HexBytes('0x00000000000000000000000094e01661ebaef430fe862f958c03200b0f483f27'),
                                          HexBytes('0xcc292d3dab2c0fbbf616670ac57ec51162959c2d9cbe938819b6e8bc1c757335')],
                               'data': '0x0000000000000000000000000000000000000000000000000000000000000001',
                               'blockNumber': 4357126,
                               'transactionHash': HexBytes('0x21381484d8f69dcd782560d1fd3cd818e743c79767985d01aec7e61c2a7f1de9'),
                               'transactionIndex': 14,
                               'blockHash': HexBytes('0x677ada1a306fc50751001bca6eeaa3f5a87a0bf2c9f6fa27899bfbaf999cca4f'),
                               'logIndex': 18,
                               'removed': False}),
                AttributeDict({'address': '0x99b9F9BA62002a9b43aF6e540428277D5E52EF47',
                               'topics': [HexBytes('0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'),
                                          HexBytes('0x00000000000000000000000099b9f9ba62002a9b43af6e540428277d5e52ef47'),
                                          HexBytes('0x00000000000000000000000094e01661ebaef430fe862f958c03200b0f483f27'),
                                          HexBytes('0xe37edda38a308a6fae15178579aab28bc7f9e46e52fde30c3a46b82b7461aa08')],
                               'data': '0x',
                               'blockNumber': 4357126,
                               'transactionHash': HexBytes('0x21381484d8f69dcd782560d1fd3cd818e743c79767985d01aec7e61c2a7f1de9'),
                               'transactionIndex': 14,
                               'blockHash': HexBytes('0x677ada1a306fc50751001bca6eeaa3f5a87a0bf2c9f6fa27899bfbaf999cca4f'),
                               'logIndex': 19,
                               'removed': False}),
                AttributeDict({'address': '0x99b9F9BA62002a9b43aF6e540428277D5E52EF47',
                               'topics': [HexBytes('0x2114851a3e2a54429989f46c1ab0743e37ded205d9bbdfd85635aed5bd595a06'),
                                          HexBytes('0x00000000000000000000000099b9f9ba62002a9b43af6e540428277d5e52ef47'),
                                          HexBytes('0x00000000000000000000000094e01661ebaef430fe862f958c03200b0f483f27'),
                                          HexBytes('0xe37edda38a308a6fae15178579aab28bc7f9e46e52fde30c3a46b82b7461aa08')],
                               'data': '0x0000000000000000000000000000000000000000000000000000000000000001',
                               'blockNumber': 4357126,
                               'transactionHash': HexBytes('0x21381484d8f69dcd782560d1fd3cd818e743c79767985d01aec7e61c2a7f1de9'),
                               'transactionIndex': 14,
                               'blockHash': HexBytes('0x677ada1a306fc50751001bca6eeaa3f5a87a0bf2c9f6fa27899bfbaf999cca4f'),
                               'logIndex': 20,
                               'removed': False}),
                AttributeDict({'address': '0x64DA772DD84965f0Ee58174941d78a9DfBccca2e',
                               'topics': [HexBytes('0x3c8fbbba495ddb1296f967c80627bcca81b77be0b349ed8ae5f604365c22e9c7')],
                               'data': '0x6e4f8d9f6517dfa28f202b2e2582943d1bd567dfb0c4a774989715c29e4aed180000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000039c4bfa00b6edeccdd00fa9589e1be76de63e862000000000000000000000000000000000000000000000003aa2371d7006800000000000000000000000000000000000000000000000000000000000000000001',
                               'blockNumber': 4357126,
                               'transactionHash': HexBytes('0x21381484d8f69dcd782560d1fd3cd818e743c79767985d01aec7e61c2a7f1de9'),
                               'transactionIndex': 14,
                               'blockHash': HexBytes('0x677ada1a306fc50751001bca6eeaa3f5a87a0bf2c9f6fa27899bfbaf999cca4f'),
                               'logIndex': 21,
                               'removed': False}),
                AttributeDict({'address': '0xe35F3B71CA90eE2606F64b645D8F4f8DCaA914Bf',
                               'topics': [HexBytes('0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef')],
                               'data': '0x0000000000000000000000008683f9c4e856be65f8a38a3a768e8fd6de94d30a000000000000000000000000fd1017c3284a12ac33bc65df12d71721c85931e00000000000000000000000000000000000000000000000000000000000000001',
                               'blockNumber': 7240557,
                               'transactionHash': HexBytes('0x1962e296457b16d5221d33623f2db5f617cb54221deb7cfd73611f761ac526a3'),
                               'transactionIndex': 6,
                               'blockHash': HexBytes('0xdf3c33d034f1b342820afdfb2612d0794d4a1c15518184d71a047ef1eb151d10'),
                               'logIndex': 5,
                               'removed': False}),
                ]

        # TODO Test with ganache
        decoded_logs = self.ethereum_client.erc20.decode_logs(logs)
        self.assertEqual(len(decoded_logs), 6)
        self.assertEqual(len([event for event in decoded_logs if 'tokenId' in event['args']]), 2)
        self.assertEqual(len([event for event in decoded_logs if 'value' in event['args']]), 3)
        self.assertEqual(len([event for event in decoded_logs if 'unknown' in event['args']]), 1)

        expected_log_1 = {'address': '0x39C4BFa00b6edecCDd00fA9589E1BE76DE63e862',
                          'topics': [HexBytes('0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'),
                                     HexBytes('0x00000000000000000000000094e01661ebaef430fe862f958c03200b0f483f27'),
                                     HexBytes('0x00000000000000000000000064da772dd84965f0ee58174941d78a9dfbccca2e')],
                          'data': '0x000000000000000000000000000000000000000000000001a055690d9db80000',
                          'blockNumber': 4357126,
                          'transactionHash': HexBytes('0x21381484d8f69dcd782560d1fd3cd818e743c79767985d01aec7e61c2a7f1de9'),
                          'transactionIndex': 14,
                          'blockHash': HexBytes('0x677ada1a306fc50751001bca6eeaa3f5a87a0bf2c9f6fa27899bfbaf999cca4f'),
                          'logIndex': 14,
                          'removed': False,
                          'args': {'from': '0x94E01661eBaef430fE862f958c03200b0F483f27',
                                   'to': '0x64DA772DD84965f0Ee58174941d78a9DfBccca2e',
                                   'value': 30000000000000000000}
                          }
        expected_log_3 = {'address': '0x99b9F9BA62002a9b43aF6e540428277D5E52EF47',
                          'topics': [HexBytes('0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'),
                                     HexBytes('0x00000000000000000000000099b9f9ba62002a9b43af6e540428277d5e52ef47'),
                                     HexBytes('0x00000000000000000000000094e01661ebaef430fe862f958c03200b0f483f27'),
                                     HexBytes('0xcc292d3dab2c0fbbf616670ac57ec51162959c2d9cbe938819b6e8bc1c757335')],
                          'data': '0x',
                          'blockNumber': 4357126,
                          'transactionHash': HexBytes('0x21381484d8f69dcd782560d1fd3cd818e743c79767985d01aec7e61c2a7f1de9'),
                          'transactionIndex': 14,
                          'blockHash': HexBytes('0x677ada1a306fc50751001bca6eeaa3f5a87a0bf2c9f6fa27899bfbaf999cca4f'), 'logIndex': 17,
                          'removed': False,
                          'args': {'from': '0x99b9F9BA62002a9b43aF6e540428277D5E52EF47',
                                   'to': '0x94E01661eBaef430fE862f958c03200b0F483f27',
                                   'tokenId': 92344574081811136966607054691835999266725413506859602442775390826469350798133}}

        expected_log_5 = {'address': '0xe35F3B71CA90eE2606F64b645D8F4f8DCaA914Bf',
                          'topics': [HexBytes('0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef')],
                          'data': '0x0000000000000000000000008683f9c4e856be65f8a38a3a768e8fd6de94d30a000000000000000000000000fd1017c3284a12ac33bc65df12d71721c85931e00000000000000000000000000000000000000000000000000000000000000001',
                          'blockNumber': 7240557,
                          'transactionHash': HexBytes(
                              '0x1962e296457b16d5221d33623f2db5f617cb54221deb7cfd73611f761ac526a3'),
                          'transactionIndex': 6,
                          'blockHash': HexBytes('0xdf3c33d034f1b342820afdfb2612d0794d4a1c15518184d71a047ef1eb151d10'),
                          'logIndex': 5,
                          'removed': False,
                          'args': {'from': '0x8683f9c4e856be65f8a38a3a768e8fd6de94d30a',
                                   'to': '0xfd1017c3284a12ac33bc65df12d71721c85931e0',
                                   'unknown': 1}
                          }

        self.assertEqual(decoded_logs[1], expected_log_1)
        self.assertEqual(decoded_logs[3], expected_log_3)
        self.assertEqual(decoded_logs[5], expected_log_5)

    def test_get_name_symbol_balance(self):
        amount = 1000
        address = Account.create().address
        erc20_contract = self.deploy_example_erc20(amount, address)
        token_balance = self.ethereum_client.erc20.get_balance(address, erc20_contract.address)
        self.assertTrue(token_balance, amount)

        another_account = Account.create().address
        token_balance = self.ethereum_client.erc20.get_balance(another_account, erc20_contract.address)
        self.assertEqual(token_balance, 0)

        self.assertTrue(self.ethereum_client.erc20.get_name(erc20_contract.address))
        self.assertTrue(self.ethereum_client.erc20.get_symbol(erc20_contract.address))

    def test_get_balances(self):
        account_address = Account.create().address
        self.assertEqual(self.ethereum_client.erc20.get_balances(account_address, []),
                         [{'token_address': None, 'balance': 0}])

        value = 7
        self.send_ether(account_address, 7)
        self.assertEqual(self.ethereum_client.erc20.get_balances(account_address, []),
                         [{'token_address': None, 'balance': value}])

        tokens_value = 12
        erc20 = self.deploy_example_erc20(tokens_value, account_address)
        self.assertCountEqual(self.ethereum_client.erc20.get_balances(account_address, [erc20.address]),
                              [{'token_address': None, 'balance': value},
                               {'token_address': erc20.address, 'balance': tokens_value}])

        tokens_value_2 = 19
        erc20_2 = self.deploy_example_erc20(tokens_value_2, account_address)
        self.assertCountEqual(self.ethereum_client.erc20.get_balances(account_address, [erc20.address,
                                                                                        erc20_2.address]),
                              [{'token_address': None, 'balance': value},
                               {'token_address': erc20.address, 'balance': tokens_value},
                               {'token_address': erc20_2.address, 'balance': tokens_value_2}
                               ])

    def test_batch_call(self):
        account_address = Account.create().address
        tokens_value = 12
        erc20 = self.deploy_example_erc20(tokens_value, account_address)
        results = self.ethereum_client.batch_call([erc20.functions.decimals(),
                                                   erc20.functions.symbol(),
                                                   erc20.functions.balanceOf(account_address)])
        decimals, symbol, balance_of = results
        self.assertEqual(decimals, erc20.functions.decimals().call())
        self.assertEqual(symbol, erc20.functions.symbol().call())
        self.assertEqual(balance_of, tokens_value)

        invalid_erc20 = get_erc20_contract(self.ethereum_client.w3, Account.create().address)
        with self.assertRaises(ValueError):
            self.ethereum_client.batch_call([invalid_erc20.functions.decimals(),
                                             invalid_erc20.functions.symbol(),
                                             invalid_erc20.functions.balanceOf(account_address)])
        # It shouldn't raise error and instead return None
        self.assertEqual(self.ethereum_client.batch_call([invalid_erc20.functions.decimals(),
                                                          invalid_erc20.functions.symbol(),
                                                          invalid_erc20.functions.balanceOf(account_address)],
                                                         raise_exception=False),
                         [None, None, None])

    def test_get_blocks(self):
        self.assertEqual(self.ethereum_client.get_blocks([]), [])
        # Generate 3 blocks
        to = Account.create().address
        value = 345
        for _ in range(3):
            self.send_ether(to, value)

        block_numbers = [self.ethereum_client.current_block_number - i for i in range(3)]
        blocks = self.ethereum_client.get_blocks(block_numbers, full_transactions=True)
        for i, block in enumerate(blocks):
            self.assertEqual(block['number'], block_numbers[i])
            self.assertEqual(len(block['hash']), 32)
            self.assertEqual(len(block['parentHash']), 32)
            self.assertGreaterEqual(len(block['transactions']), 0)

    def test_get_total_transfer_history(self):
        amount = 50
        owner_account = self.create_account(initial_ether=0.01)
        account_1 = self.create_account(initial_ether=0.01)
        account_2 = self.create_account(initial_ether=0.01)
        account_3 = self.create_account(initial_ether=0.01)
        erc20_contract = self.deploy_example_erc20(amount, owner_account.address)
        # `owner` sends `amount // 2` to `account_1` and `account_3`
        self.send_tx(erc20_contract.functions.transfer(account_1.address,
                                                       amount // 2).buildTransaction({'from': owner_account.address}),
                     owner_account)
        self.send_tx(erc20_contract.functions.transfer(account_3.address,
                                                       amount // 2).buildTransaction({'from': owner_account.address}),
                     owner_account)
        logs = self.ethereum_client.erc20.get_total_transfer_history([account_1.address])
        self.assertEqual(len(logs), 1)
        self.assertEqual(logs[0]['args']['from'], owner_account.address)
        self.assertEqual(logs[0]['args']['to'], account_1.address)
        self.assertEqual(logs[0]['args']['value'], amount // 2)
        log_0_block_number = logs[0]['blockNumber']

        # `account1` sends `amount // 2` (all) to `account_2`
        self.send_tx(erc20_contract.functions.transfer(account_2.address,
                                                       amount // 2).buildTransaction({'from': account_1.address}),
                     account_1)
        # Test `token_address` and `to_block` parameters
        logs = self.ethereum_client.erc20.get_total_transfer_history([account_1.address],
                                                                     token_address=erc20_contract.address,
                                                                     to_block='latest')
        self.assertEqual(len(logs), 2)
        self.assertEqual(logs[1]['args']['from'], account_1.address)
        self.assertEqual(logs[1]['args']['to'], account_2.address)
        self.assertEqual(logs[1]['args']['value'], amount // 2)
        self.assertGreaterEqual(logs[1]['blockNumber'], log_0_block_number)

        logs = self.ethereum_client.erc20.get_total_transfer_history([account_3.address])
        self.assertEqual(len(logs), 1)
        self.assertEqual(logs[0]['args']['from'], owner_account.address)
        self.assertEqual(logs[0]['args']['to'], account_3.address)
        self.assertEqual(logs[0]['args']['value'], amount // 2)

        logs = self.ethereum_client.erc20.get_total_transfer_history([account_2.address, account_3.address])
        self.assertEqual(len(logs), 2)

        logs = self.ethereum_client.erc20.get_total_transfer_history()
        self.assertGreaterEqual(len(logs), 3)

    def test_get_transfer_history(self):
        amount = 1000
        owner_account = self.create_account(initial_ether=0.01)

        # Owner will send amount / 2 to receiver and receiver2. Then receiver1 and receiver 2
        # will send amount / 4 to receiver3
        receiver_account = self.create_account(initial_ether=0.01)
        receiver2_account = self.create_account(initial_ether=0.01)
        receiver3_account = self.create_account(initial_ether=0.01)
        erc20_contract = self.deploy_example_erc20(amount, owner_account.address)
        block_number = self.w3.eth.block_number
        events = self.ethereum_client.erc20.get_transfer_history(block_number, token_address=erc20_contract.address)
        self.assertEqual(len(events), 1)  # Event is triggered on minting

        self.send_tx(erc20_contract.functions.transfer(receiver_account.address,
                                                       amount // 2).buildTransaction({'from': owner_account.address}),
                     owner_account)
        self.send_tx(erc20_contract.functions.transfer(receiver2_account.address,
                                                       amount // 2).buildTransaction({'from': owner_account.address}),
                     owner_account)

        self.send_tx(erc20_contract.functions.transfer(receiver3_account.address,
                                                       amount // 4).buildTransaction({'from': receiver_account.address}),
                     receiver_account)
        self.send_tx(erc20_contract.functions.transfer(receiver3_account.address,
                                                       amount // 4).buildTransaction({'from': receiver2_account.address}),
                     receiver2_account)

        events = self.ethereum_client.erc20.get_transfer_history(block_number, token_address=erc20_contract.address)
        self.assertEqual(len(events), 5)

        events = self.ethereum_client.erc20.get_transfer_history(block_number, from_address=owner_account.address)
        self.assertEqual(len(events), 2)

        events = self.ethereum_client.erc20.get_transfer_history(block_number, from_address=receiver_account.address)
        self.assertEqual(len(events), 1)

        events = self.ethereum_client.erc20.get_transfer_history(block_number, from_address=receiver2_account.address)
        self.assertEqual(len(events), 1)

        events = self.ethereum_client.erc20.get_transfer_history(block_number, from_address=receiver3_account.address)
        self.assertEqual(len(events), 0)

        events = self.ethereum_client.erc20.get_transfer_history(block_number, to_address=receiver2_account.address)
        self.assertEqual(len(events), 1)

        events = self.ethereum_client.erc20.get_transfer_history(block_number, to_address=receiver3_account.address)
        self.assertEqual(len(events), 2)
        for event in events:
            self.assertEqual(event.args.value, amount // 4)
            self.assertEqual(event.args.to, receiver3_account.address)

        events = self.ethereum_client.erc20.get_transfer_history(block_number,
                                                                 from_address=receiver2_account.address,
                                                                 to_address=receiver3_account.address)
        self.assertEqual(len(events), 1)
        event = events[0]
        self.assertEqual(event.args.value, amount // 4)
        self.assertEqual(event.args['from'], receiver2_account.address)
        self.assertEqual(event.args.to, receiver3_account.address)

    def test_get_info(self):
        amount = 1
        owner = Account.create().address
        erc20_contract = self.deploy_example_erc20(amount, owner)
        erc20_info = self.ethereum_client.erc20.get_info(erc20_contract.address)
        self.assertEqual(erc20_info.decimals, 18)

        with self.assertRaises(InvalidERC20Info):
            self.ethereum_client.erc20.get_info(Account.create().address)

    def test_send_tokens(self):
        amount = 5
        owner = self.ethereum_test_account
        owner_2 = Account.create()
        erc20 = self.deploy_example_erc20(amount, owner.address)
        self.assertEqual(self.ethereum_client.erc20.get_balance(owner.address, erc20.address), amount)

        amount_2 = 3
        self.ethereum_client.erc20.send_tokens(owner_2.address, amount_2, erc20.address, owner.key)
        self.assertEqual(self.ethereum_client.erc20.get_balance(owner.address, erc20.address), amount - amount_2)
        self.assertEqual(self.ethereum_client.erc20.get_balance(owner_2.address, erc20.address), amount_2)


class TestParityManager(EthereumTestCaseMixin, TestCase):
    def test_decode_trace(self):
        example_traces = [
            {
                "action": {
                    "callType": "call",
                    "from": "0x32be343b94f860124dc4fee278fdcbd38c102d88",
                    "gas": "0x4c40d",
                    "input": "0x",
                    "to": "0x8bbb73bcb5d553b5a556358d27625323fd781d37",
                    "value": "0x3f0650ec47fd240000"
                },
                "blockHash": "0x86df301bcdd8248d982dbf039f09faf792684e1aeee99d5b58b77d620008b80f",
                "blockNumber": 3068183,
                "result": {
                    "gasUsed": "0x0",
                    "output": "0x"
                },
                "subtraces": 0,
                "traceAddress": [],
                "transactionHash": "0x3321a7708b1083130bd78da0d62ead9f6683033231617c9d268e2c7e3fa6c104",
                "transactionPosition": 3,
                "type": "call"
            },
            {
                "action": {
                    "from": "0x3b169a0fb55ea0b6bafe54c272b1fe4983742bf7",
                    "gas": "0x49b0b",
                    "init": " ",
                    "value": "0x0"
                },
                "blockHash": "0x03f9f64dfeb7807b5df608e6957dd4d521fd71685aac5533451d27f0abe03660",
                "blockNumber": 3793534,
                "result": {
                    "address": "0x61a7cc907c47c133d5ff5b685407201951fcbd08",
                    "code": "0x60806040526004361061004c576000357c0100000000000000000000000000000000000000000000000000000000900463ffffffff1680634555d5c91461008b5780635c60da1b146100b6575b73ffffffffffffffffffffffffffffffffffffffff600054163660008037600080366000845af43d6000803e6000811415610086573d6000fd5b3d6000f35b34801561009757600080fd5b506100a061010d565b6040518082815260200191505060405180910390f35b3480156100c257600080fd5b506100cb610116565b604051808273ffffffffffffffffffffffffffffffffffffffff1673ffffffffffffffffffffffffffffffffffffffff16815260200191505060405180910390f35b60006002905090565b60008060009054906101000a900473ffffffffffffffffffffffffffffffffffffffff169050905600a165627a7a7230582007fffd557dfc8c4d2fdf56ba6381a6ce5b65b6260e1492d87f26c6d4f1d041080029",
                    "gasUsed": "0x4683f"
                },
                "subtraces": 2,
                "traceAddress": [],
                "transactionHash": "0x6c7e8f8778d33d81b29c4bd7526ee50a4cea340d69eed6c89ada4e6fab731789",
                "transactionPosition": 1,
                "type": "create"
            },
            {
                "action": {
                    "callType": "call",
                    "from": "0xed85033a23027439fb20f381f5930ba67f1ebee0",
                    "gas": "0x0",
                    "input": "0x",
                    "to": "0x0affccd762f0a7fa1bca40c51afe1a806a74a6f9",
                    "value": "0x11c290633b7000"
                },
                "blockHash": "0xbf3a508137e72b20399852becdee1752c7b9986184f835014d12b10c01746f39",
                "blockNumber": 7556887,
                "error": "Out of gas",
                "subtraces": 0,
                "traceAddress": [],
                "transactionHash": "0xec7e447ce8eef033a8c85442a281bd34436e576553c7d98fbe859af7754a9064",
                "transactionPosition": 117,
                "type": "call"
            },
            {'action': {'address': '0x4440adafbc6c4e45c299451c0eedc7c8b98c14ac',
                        'balance': '0xa',
                        'refundAddress': '0x1240adafbc6c4e45c299451c0eedc7c8b98c2222'},
             'blockHash': '0x8512d367492371edf44ebcbbbd935bc434946dddc2b126cb558df5906012186c',
             'blockNumber': 7829689,
             'result': None,
             'subtraces': 0,
             'traceAddress': [0, 0, 0, 0, 0, 0],
             'transactionHash': '0x5f7af6aa390f9f8dd79ee692c37cbde76bb7869768b1bac438b6d176c94f637d',
             'transactionPosition': 35,
             'type': 'suicide'},
        ]
        at_least_one_error = False
        at_least_one_self_destruct = False
        decoded_traces = self.ethereum_client.parity._decode_traces(example_traces)
        for example_trace, decoded_trace in zip(example_traces, decoded_traces):
            if decoded_trace['type'] == 'suicide':
                self.assertEqual(decoded_trace['action']['address'], '0x4440AdaFBc6c4E45C299451C0eEdC7C8B98c14Ac')
                self.assertEqual(decoded_trace['action']['balance'], 10)
                self.assertEqual(decoded_trace['action']['refundAddress'], '0x1240aDafBC6C4e45C299451C0eEdC7c8B98C2222')
                at_least_one_self_destruct = True
            else:
                self.assertEqual(decoded_trace['action']['gas'], int(example_trace['action']['gas'], 16))
                self.assertEqual(decoded_trace['action']['value'], int(example_trace['action']['value'], 16))
                if 'error' in decoded_trace:
                    self.assertNotIn('result', decoded_trace)
                    at_least_one_error = True
                else:
                    self.assertEqual(decoded_trace['result']['gasUsed'], int(example_trace['result']['gasUsed'], 16))

        self.assertTrue(at_least_one_error)
        self.assertTrue(at_least_one_self_destruct)

        self.assertEqual(decoded_traces[0]['result']['output'], HexBytes(''))
        self.assertEqual(decoded_traces[1]['result']['address'],
                         self.w3.toChecksumAddress(example_traces[1]['result']['address']))
        self.assertEqual(decoded_traces[1]['result']['code'],
                         HexBytes(example_traces[1]['result']['code']))

        self.assertEqual(decoded_traces[2]['error'], 'Out of gas')

    def test_filter_out_errored_traces(self):
        self.assertEqual(self.ethereum_client.parity.filter_out_errored_traces([]), [])
        traces = internal_txs_errored
        expected = [
            {'action': {'from': '0x667dEb5A98f77052cf561658575cF1530Ee42C7a',
                        'gas': 60066,
                        'value': 0,
                        'callType': 'call',
                        'input': HexBytes(' '),
                        'to': '0x47F61944efdB020829caead65AfF8AC024600580'},
             'blockHash': '0x4c49052fc99be82b91f8a35320826304dfe278dbd7d756edde000d331606358f',
             'blockNumber': 4735890,
             'result': {'gasUsed': 21838,
                        'output': HexBytes('0x0000000000000000000000000000000000000000000000000000000000000000')},
             'subtraces': 1,
             'traceAddress': [],
             'transactionHash': '0xf097d5e5dd39a6799fc13dfa49732a115b457386520dc92f99f0135a1d196851',
             'transactionPosition': 3,
             'type': 'call'},
            {'action': {'from': '0x47F61944efdB020829caead65AfF8AC024600580',
                        'gas': 57726,
                        'value': 0,
                        'callType': 'delegatecall',
                        'input': HexBytes(' '),
                        'to': '0xb6029EA3B2c51D09a50B53CA8012FeEB05bDa35A'},
             'blockHash': '0x4c49052fc99be82b91f8a35320826304dfe278dbd7d756edde000d331606358f',
             'blockNumber': 4735890,
             'result': {'gasUsed': 20369,
                        'output': HexBytes('0x0000000000000000000000000000000000000000000000000000000000000000')},
             'subtraces': 1,
             'traceAddress': [0],
             'transactionHash': '0xf097d5e5dd39a6799fc13dfa49732a115b457386520dc92f99f0135a1d196851',
             'transactionPosition': 3,
             'type': 'call'}]
        self.assertEqual(self.ethereum_client.parity.filter_out_errored_traces(traces), expected)

        traces = [
            {
                'traceAddress': [],
            },
            {
                'traceAddress': [0],
            },
            {
                'traceAddress': [0, 0],
            },
            {
                'error': 'reverted',
                'traceAddress': [0, 0, 0],
            },
            {
                'traceAddress': [0, 0, 0, 0],
            },
            {
                'traceAddress': [0, 0, 0, 1],
            },
            {
                'traceAddress': [0, 0, 1],
            },
            {
                'traceAddress': [0, 0, 1, 0],
            },
            {
                'error': 'reverted',
                'traceAddress': [0, 0, 1, 0, 0],
            },
            {
                'traceAddress': [0, 0, 1, 0, 1],
            },
            {
                'traceAddress': [0, 0, 1, 0, 2],
            },
            {
                'error': 'reverted',
                'traceAddress': [0, 0, 1, 0, 3],
            },
            {
                'traceAddress': [0, 0, 2],
            },
        ]
        expected = [
            {
                'traceAddress': [],
            },
            {
                'traceAddress': [0],
            },
            {
                'traceAddress': [0, 0],
            },
            {
                'traceAddress': [0, 0, 1],
            },
            {
                'traceAddress': [0, 0, 1, 0],
            },
            {
                'traceAddress': [0, 0, 1, 0, 1],
            },
            {
                'traceAddress': [0, 0, 1, 0, 2],
            },
            {
                'traceAddress': [0, 0, 2],
            },
        ]
        self.assertEqual(sorted(traces, key=lambda trace: trace['traceAddress']), traces)
        self.assertEqual(self.ethereum_client.parity.filter_out_errored_traces(traces), expected)

    @mock.patch.object(ParityManager, 'trace_transaction', return_value=internal_txs_errored)
    def test_get_previous_trace(self, trace_transaction_mock: MagicMock):
        self.assertEqual(self.ethereum_client.parity.get_previous_trace('0x12', [0, 0])['traceAddress'],
                         [0])
        self.assertEqual(self.ethereum_client.parity.get_previous_trace('0x12', [0, 0],
                                                                        number_traces=2)['traceAddress'],
                         [])
        self.assertEqual(self.ethereum_client.parity.get_previous_trace('0x12', [0, 0],
                                                                        skip_delegate_calls=True)['traceAddress'],
                         [])
        self.assertIsNone(self.ethereum_client.parity.get_previous_trace('0x12', [0, 0],
                                                                         number_traces=3))
        self.assertEqual(self.ethereum_client.parity.get_previous_trace('0x12', [0, 0, 0],
                                                                        skip_delegate_calls=True)['traceAddress'],
                         [0, 0])

    @mock.patch.object(ParityManager, 'trace_transaction', return_value=creation_internal_txs)
    def test_get_next_traces(self, trace_transaction_mock: MagicMock):
        def trace_addresses(traces: Sequence[Dict[str, Any]]) -> List[List[int]]:
            return [trace['traceAddress'] for trace in traces]

        self.assertEqual(trace_addresses(self.ethereum_client.parity.get_next_traces('0x12', [])), [[0], [1]])
        self.assertEqual(trace_addresses(self.ethereum_client.parity.get_next_traces('0x12', [],
                                                                                     remove_delegate_calls=True)),
                         [[1]])
        self.assertEqual(trace_addresses(self.ethereum_client.parity.get_next_traces('0x12', [], remove_calls=True)),
                         [[0]])
        self.assertEqual(trace_addresses(self.ethereum_client.parity.get_next_traces('0x12', [],
                                                                                     remove_delegate_calls=True,
                                                                                     remove_calls=True)),
                         [])
        self.assertEqual(self.ethereum_client.parity.get_next_traces('0x12', [0]), [])
        self.assertEqual(trace_addresses(self.ethereum_client.parity.get_next_traces('0x12', [1])), [[1, 0]])

    def test_trace_filter(self):
        with self.assertRaisesMessage(AssertionError, 'at least'):
            self.ethereum_client.parity.trace_filter()

        with self.assertRaisesMessage(ValueError, 'Method trace_filter not supported'):
            self.ethereum_client.parity.trace_filter(from_address=Account.create().address)


class TestEthereumNetwork(EthereumTestCaseMixin, TestCase):
    def test_default_ethereum_network_name(self):
        self.assertEqual(EthereumNetwork(EthereumNetwork.default), EthereumNetwork.UNKNOWN)

    def test_unknown_ethereum_network_name(self):
        self.assertEqual(EthereumNetwork(2), EthereumNetwork.UNKNOWN)

    def test_mainnet_ethereum_network_name(self):
        self.assertEqual(EthereumNetwork(1), EthereumNetwork.MAINNET)

    def test_rinkeby_ethereum_network_name(self):
        self.assertEqual(EthereumNetwork(4), EthereumNetwork.RINKEBY)


class TestEthereumClient(EthereumTestCaseMixin, TestCase):
    def test_ethereum_client_str(self):
        self.assertTrue(str(self.ethereum_client))

    def test_current_block_number(self):
        self.assertGreaterEqual(self.ethereum_client.current_block_number, 0)

    def test_get_transaction(self):
        to = Account.create().address
        value = 123
        tx_hash = self.send_ether(to, value)
        tx = self.ethereum_client.get_transaction(tx_hash)
        self.assertEqual(tx.to, to)
        self.assertEqual(tx.value, value)
        block = self.ethereum_client.get_block(tx.blockNumber)
        self.assertEqual(block.number, tx.blockNumber)

    def test_get_transactions(self):
        self.assertEqual(self.ethereum_client.get_transactions([]), [])
        to = Account.create().address
        values = [123, 234, 567]
        tx_hashes = [self.send_ether(to, values[i]) for i in range(3)]
        txs = self.ethereum_client.get_transactions(tx_hashes)
        for i, tx in enumerate(txs):
            self.assertEqual(tx['to'], to)
            self.assertEqual(tx['value'], values[i])

        self.assertEqual(self.ethereum_client.get_transaction_receipts([]), [])
        receipts = self.ethereum_client.get_transaction_receipts(tx_hashes)
        for i, receipt in enumerate(receipts):
            self.assertEqual(receipt['status'], 1)
            self.assertGreaterEqual(receipt['gasUsed'], 21000)

    def test_check_tx_with_confirmations(self):
        value = 1
        to, _ = get_eth_address_with_key()

        tx_hash = self.ethereum_client.send_eth_to(self.ethereum_test_account.key,
                                                   to=to, gas_price=self.gas_price, value=value)
        self.assertFalse(self.ethereum_client.check_tx_with_confirmations(tx_hash, 2))

        _ = self.ethereum_client.send_eth_to(self.ethereum_test_account.key,
                                             to=to, gas_price=self.gas_price, value=value)
        self.assertFalse(self.ethereum_client.check_tx_with_confirmations(tx_hash, 2))

        _ = self.ethereum_client.send_eth_to(self.ethereum_test_account.key,
                                             to=to, gas_price=self.gas_price, value=value)
        self.assertTrue(self.ethereum_client.check_tx_with_confirmations(tx_hash, 2))

    def test_estimate_gas(self):
        send_ether_gas = 21000
        from_ = self.ethereum_test_account.address
        to, _ = get_eth_address_with_key()
        gas = self.ethereum_client.estimate_gas(to, from_=from_, value=5, data=None, block_identifier='pending')
        self.assertEqual(gas, send_ether_gas)

        gas = self.ethereum_client.estimate_gas(to, from_=from_, value=5, data=b'')
        self.assertEqual(gas, send_ether_gas)

        gas = self.ethereum_client.estimate_gas(to, from_=from_, value=5, data=None)
        self.assertEqual(gas, send_ether_gas)

    def test_estimate_gas_with_erc20(self):
        send_ether_gas = 21000
        amount_tokens = 5000
        amount_to_send = amount_tokens // 2
        from_account = self.ethereum_test_account
        from_ = from_account.address
        to, _ = get_eth_address_with_key()

        erc20_contract = self.deploy_example_erc20(amount_tokens, from_)
        transfer_tx = erc20_contract.functions.transfer(to, amount_to_send).buildTransaction({'from': from_})
        data = transfer_tx['data']

        # Ganache does not cares about all this anymore
        # ---------------------------------------------
        # Ganache returns error because there is data but address is not a contract
        # with self.assertRaisesMessage(ValueError, 'transaction which calls a contract function'):
        #    self.ethereum_client.estimate_gas(from_, to, 0, data, block_identifier='pending')
        # `to` is not a contract, no functions will be triggered
        # gas = self.ethereum_client.estimate_gas(from_, to, 0, data, block_identifier='pending')
        # self.assertGreater(gas, send_ether_gas)
        # self.assertLess(gas, send_ether_gas + 2000)

        gas = self.ethereum_client.estimate_gas(erc20_contract.address, from_=from_, value=0, data=data,
                                                block_identifier='pending')
        self.assertGreater(gas, send_ether_gas)

        # We do the real erc20 transfer for the next test case
        self.send_tx(transfer_tx, from_account)
        token_balance = self.ethereum_client.erc20.get_balance(to, erc20_contract.address)
        self.assertTrue(token_balance, amount_to_send)

        # Gas will be lowest now because you don't have to pay for storage
        gas2 = self.ethereum_client.estimate_gas(erc20_contract.address, from_=from_, value=0, data=data,
                                                 block_identifier='pending')
        self.assertLess(gas2, gas)

    def test_get_ethereum_network(self):
        self.assertEqual(self.ethereum_client.get_network(), EthereumNetwork.GANACHE)

        with mock.patch.object(Net, 'version', return_value='1', new_callable=mock.PropertyMock):
            self.assertEqual(self.ethereum_client.get_network(), EthereumNetwork.MAINNET)

        with mock.patch.object(Net, 'version', return_value='4', new_callable=mock.PropertyMock):
            self.assertEqual(self.ethereum_client.get_network(), EthereumNetwork.RINKEBY)

        with mock.patch.object(Net, 'version', return_value='4815162342', new_callable=mock.PropertyMock):
            self.assertEqual(self.ethereum_client.get_network(), EthereumNetwork.UNKNOWN)

    def test_get_nonce(self):
        address = Account.create().address
        nonce = self.ethereum_client.get_nonce_for_account(address)
        self.assertEqual(nonce, 0)

        nonce = self.ethereum_client.get_nonce_for_account(address, block_identifier='pending')
        self.assertEqual(nonce, 0)

    def test_estimate_data_gas(self):
        self.assertEqual(self.ethereum_client.estimate_data_gas(HexBytes('')), 0)
        self.assertEqual(self.ethereum_client.estimate_data_gas(HexBytes('0x00')), 4)
        self.assertEqual(self.ethereum_client.estimate_data_gas(HexBytes('0x000204')), 4 + GAS_CALL_DATA_BYTE * 2)
        self.assertEqual(self.ethereum_client.estimate_data_gas(HexBytes('0x050204')), GAS_CALL_DATA_BYTE * 3)
        self.assertEqual(self.ethereum_client.estimate_data_gas(HexBytes('0x0502040000')), GAS_CALL_DATA_BYTE * 3 + 4 * 2)
        self.assertEqual(self.ethereum_client.estimate_data_gas(HexBytes('0x050204000001')), GAS_CALL_DATA_BYTE * 4 + 4 * 2)
        self.assertEqual(self.ethereum_client.estimate_data_gas(HexBytes('0x00050204000001')),
                         4 + GAS_CALL_DATA_BYTE * 4 + 4 * 2)

    def test_provider_singleton(self):
        ethereum_client1 = EthereumClientProvider()
        ethereum_client2 = EthereumClientProvider()
        self.assertEqual(ethereum_client1, ethereum_client2)

    def test_send_eth_to(self):
        address, _ = get_eth_address_with_key()
        value = 1
        self.ethereum_client.send_eth_to(self.ethereum_test_account.key, address, self.gas_price, value)
        self.assertEqual(self.ethereum_client.get_balance(address), value)

    def test_send_eth_without_key(self):
        with self.settings(SAFE_FUNDER_PRIVATE_KEY=None):
            self.test_send_eth_to()

    def test_send_transaction(self):
        account = self.ethereum_test_account
        address = account.address
        to, _ = get_eth_address_with_key()
        value = 1
        tx = {
            'to': to,
            'value': value,
            'gas': 23000,
            'gasPrice': self.gas_price,
            'nonce': self.ethereum_client.get_nonce_for_account(address)
        }

        with self.assertRaises(FromAddressNotFound):
            self.ethereum_client.send_transaction(tx)

        # Random address not in the node
        tx['from'] = Account.create().address
        with self.assertRaises(SenderAccountNotFoundInNode):
            self.ethereum_client.send_transaction(tx)

        tx['from'] = address
        self.ethereum_client.send_transaction(tx)
        self.assertEqual(self.ethereum_client.get_balance(to), 1)

        with self.assertRaises(InvalidNonce):
            self.ethereum_client.send_transaction(tx)

        tx['value'] = self.ethereum_client.get_balance(address) + 1
        tx['nonce'] = self.ethereum_client.get_nonce_for_account(address)
        with self.assertRaises(InsufficientFunds):
            self.ethereum_client.send_transaction(tx)

    def test_send_unsigned_transaction(self):
        account = self.ethereum_test_account
        address = account.address
        to, _ = get_eth_address_with_key()
        value = 4

        tx = {
            'to': to,
            'value': value,
            'gas': 23000,
            'gasPrice': 1,
        }

        with self.assertRaisesMessage(ValueError, 'Ethereum account was not configured'):
            self.ethereum_client.send_unsigned_transaction(tx)

        self.ethereum_client.send_unsigned_transaction(tx, public_key=address)
        self.assertEqual(self.ethereum_client.get_balance(to), value)
        first_nonce = tx['nonce']
        self.assertGreaterEqual(first_nonce, 0)

        # Will use the same nonce
        with self.assertRaisesMessage(InvalidNonce, 'correct nonce'):
            self.ethereum_client.send_unsigned_transaction(tx, public_key=address)

        # With retry, everything should work
        self.ethereum_client.send_unsigned_transaction(tx, public_key=address, retry=True)
        self.assertEqual(tx['nonce'], first_nonce + 1)
        self.assertEqual(self.ethereum_client.get_balance(to), value * 2)

        # We try again with the first nonce, and should work too
        tx['nonce'] = first_nonce
        self.ethereum_client.send_unsigned_transaction(tx, public_key=address, retry=True)
        self.assertEqual(tx['nonce'], first_nonce + 2)
        self.assertEqual(self.ethereum_client.get_balance(to), value * 3)

    def test_send_unsigned_transaction_with_private_key(self):
        account = self.create_account(initial_ether=0.1)
        key = account.key
        to, _ = get_eth_address_with_key()
        value = 4

        tx = {
            'to': to,
            'value': value,
            'gas': 23000,
            'gasPrice': self.gas_price,
        }

        self.ethereum_client.send_unsigned_transaction(tx, private_key=key)
        self.assertEqual(self.ethereum_client.get_balance(to), value)
        first_nonce = tx['nonce']
        self.assertGreaterEqual(first_nonce, 0)

        # Will use the same nonce
        with self.assertRaisesMessage(InvalidNonce, 'correct nonce'):
            self.ethereum_client.send_unsigned_transaction(tx, private_key=key)

        # With retry, everything should work
        self.ethereum_client.send_unsigned_transaction(tx, private_key=key, retry=True)
        self.assertEqual(tx['nonce'], first_nonce + 1)
        self.assertEqual(self.ethereum_client.get_balance(to), value * 2)

        # We try again with the first nonce, and should work too
        tx['nonce'] = first_nonce
        self.ethereum_client.send_unsigned_transaction(tx, private_key=key, retry=True)
        self.assertEqual(tx['nonce'], first_nonce + 2)
        self.assertEqual(self.ethereum_client.get_balance(to), value * 3)

    def test_wait_for_tx_receipt(self):
        value = 1
        to = Account.create().address

        tx_hash = self.ethereum_client.send_eth_to(self.ethereum_test_account.key,
                                                   to=to, gas_price=self.gas_price, value=value)
        receipt1 = self.ethereum_client.get_transaction_receipt(tx_hash, timeout=None)
        receipt2 = self.ethereum_client.get_transaction_receipt(tx_hash, timeout=20)
        self.assertIsNotNone(receipt1)
        self.assertEqual(receipt1, receipt2)

        fake_tx_hash = self.w3.keccak(0)
        self.assertIsNone(self.ethereum_client.get_transaction_receipt(fake_tx_hash, timeout=None))
        self.assertIsNone(self.ethereum_client.get_transaction_receipt(fake_tx_hash, timeout=1))
